/*  $Id: bldCvtTable.c,v 1.15 1995/03/02 18:50:35 jba Exp $
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	09-18-91	jba	Fixed bug in break point conversion for small,<10 element,tables
 * .02  10-28-91        rcz     Removed dimension requirement
 * .03  03-28-92        jba     Fixed starting point for create breakpt table algorithm
 * .04  04-06-92        rcz     Changed input file format (cvtHeader.ascii) per mrk
 *                              Allow selections w/o any break point table (.data file)
 * .05  05-18-92        rcz     cleanup for mrk - New database access
 * .06  08-30-93        rcz     added "inline breakpoints" and changed sll to ell lists
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <choice.h>
#include <dbStaticLib.h>
#include <cvtTable.h>
#include <sdrHeader.h>
#include <getSdrLine.h>
#include <ellLib.h>

#define CTRL_TYPE1 0 /* "breakNameString" only */
#define CTRL_TYPE2 1 /* "breakNameString" plus 8 fields*/
#define CTRL_TYPE3 2 /* "breakNameString" followed by "BreakTable" keyword */

/****************************************************************************/
/* These break point name indices are reserved and hard coded at the present time 4/92 */
char *brkNames [] = {
	"NO CONVERSION", /* slot 0 */
	"LINEAR", /* slot 1 */
};
#define MAX_RES_CTRL_LINES (sizeof(brkNames) / sizeof(void *))
/****************************************************************************/
#define BYTES50		50
#define MAX_BREAKS 100
#define SAME 0
char            inFile[BYTES50];
char            choiceSdr[BYTES50];
char            cvtSdr[BYTES50];
struct sdrHeader sdrHeader;
struct arrBrkTable *cvtTable=NULL;
struct choiceSet *choiceCvt=NULL;

static struct inLineInfo {	/* temp storage for getSdrLine input fields */
    char            brkName[BYTES50];
}               inLineInfo;

typedef struct enode {
    ELLNODE          node;
    int             index_num;
    char            brkName[BYTES50];
}               ENODE;

typedef struct fnode {
    ELLNODE          node;
	long		col1;
	float		col2;
}               FNODE;

struct brkCreateInfo {
    float           engLow;	/* Lowest value desired: engineering units */
    float           engHigh;	/* Highest value desired: engineering units */
    long            rawLow;	/* Raw value for EngLow			 */
    long            rawHigh;	/* Raw value for EngHigh		 */
    float           accuracy;	/* accuracy desired in engineering units */
    float           tblEngFirst;/* First table value: engineering units */
    float           tblEngLast;	/* Last table value: engineering units	 */
    float           tblEngDelta;/* Change per table entry: eng units	 */
    long            nTable;	/* number of table entries 		 */
    /* (last-first)/delta + 1		 */
    float          *pTable;	/* addr of data table			 */
}               brkCreateInfo;
struct brkInt   brkInt[MAX_BREAKS];

static int create_break();
static int read_data();
static int appendIndexToList();
static int cmd_CvtTable();


/*
 * BLDCVTTABLE
 *
 * builds SDR files
 *
 * INPUT:	cvtTable.ascii
 *
 * OUTPUT:	choiceCvt.sdr
 * 		cvtTable.sdr
 */

main(argc, argv)
    int             argc;
    char          **argv;
{
    FILE           *inFile_fp;
    FILE           *choiceSdr_fp;
    FILE           *cvtTable_fp;
    ELLLIST          *pList = NULL;
    struct brkCreateInfo *pbrk = &brkCreateInfo;
    struct lineIn  *plineIn = NULL;
    struct inLineInfo *pInfo = &inLineInfo;
    struct arrBrkTable *pTcvtTable = NULL;
    struct choiceSet *pTchoiceCvt = NULL;
    long            Size1 = 0;
    long            SSize1 = 0;
    long            Size2 = 0;
    long            SSize2 = 0;
    void            *addr1;
    void            *addr2;
    void            *addrS1;
    void            *addrS2;
    int             i;
    int             j;
    int             rtnval;
    int             len;
    int             max_inx;
    int             max_brk_tables;


    /* check cmd line */
    if ((cmd_CvtTable(argc, argv)) < 0) {
	return (-1);
    }
    /* allocate space for the getSdrLine input structure */
    plineIn = dbCalloc(1,sizeof(struct lineIn));
    plineIn->fname = inFile;

    if ((inFile_fp = fopen(plineIn->fname, "r")) == NULL) {
	printf("bldCvtTable: fopen error errno=%i\nfile=%s\n", errno, plineIn->fname);
	return (-1);

    }
    /* allocate index list */
    pList = dbCalloc(1,sizeof(ELLLIST));
    /* 1st pass - determine number of brk tables */
    if ((bldIndexList(inFile_fp, plineIn, pList, pInfo, pbrk, &max_inx)) < 0) {
	return (-1);
    }
    rewind(inFile_fp);
    max_brk_tables = max_inx + 1;
    Size1 = sizeof(struct arrBrkTable) + max_brk_tables  * sizeof(void *);
    pTcvtTable = dbCalloc(1,Size1);
    pTcvtTable->number = max_brk_tables;
    pTcvtTable->papBrkTable =
	(void *)((char *)pTcvtTable + sizeof(struct arrBrkTable));

    Size2 = sizeof(struct choiceSet) + max_brk_tables * sizeof(void *);
    pTchoiceCvt = dbCalloc(1,Size2);
    pTchoiceCvt->number = max_brk_tables;
    pTchoiceCvt->papChoice =
	(void *)((char *)pTchoiceCvt + sizeof(struct choiceSet));

    if ((buildTables(inFile_fp, plineIn, pInfo, pbrk, pTcvtTable,
		     pTchoiceCvt,
		     &Size1,
		     &SSize1,
		     &SSize2)) < 0) {
	return (-1);
    }
    /* allocate, init and write the contiguous area for cvtTable.sdr */
    SSize1 = (SSize1 + sizeof(long) - 1) / sizeof(long) * sizeof(long);
    Size1 += SSize1;
    cvtTable = dbCalloc(1,Size1);
    addr1 = (char *)cvtTable + sizeof(struct arrBrkTable);
    cvtTable->number = pTcvtTable->number;
    cvtTable->papBrkTable = addr1;
    addr1 = (char *)addr1 + cvtTable->number * sizeof(long);
    addrS1 = (char *)cvtTable + Size1 - SSize1;
    for (i = 0; i < pTcvtTable->number; i++) {
	if (pTcvtTable->papBrkTable[i] != NULL) {
	    cvtTable->papBrkTable[i] = addr1;
	    addr1 = (char *)addr1 + sizeof(struct brkTable);
	    cvtTable->papBrkTable[i]->name = addrS1;
	    len = strlen(pTcvtTable->papBrkTable[i]->name);
	    memcpy(addrS1, pTcvtTable->papBrkTable[i]->name, len + 1);
	    cvtTable->papBrkTable[i]->number = pTcvtTable->papBrkTable[i]->number;
	    cvtTable->papBrkTable[i]->rawLow = pTcvtTable->papBrkTable[i]->rawLow;
	    cvtTable->papBrkTable[i]->rawHigh = pTcvtTable->papBrkTable[i]->rawHigh;
	    cvtTable->papBrkTable[i]->papBrkInt = addr1;
	    addr1 = (char *)addr1 + cvtTable->papBrkTable[i]->number * sizeof(void *);
	    /* note: always a minimum of two break points */
	    for (j = 0; j < pTcvtTable->papBrkTable[i]->number; j++) {
		cvtTable->papBrkTable[i]->papBrkInt[j] = addr1;
		memcpy((char *)cvtTable->papBrkTable[i]->papBrkInt[j],
			(char *)pTcvtTable->papBrkTable[i]->papBrkInt[j],
			sizeof(struct brkInt));
		addr1 = (char *)addr1 + sizeof(struct brkInt);
	    }
	    addrS1 = (char *)addrS1 + strlen(cvtTable->papBrkTable[i]->name) + 1;
	}
    }
    /* make all pointers offsets */
    for (i = 0; i < cvtTable->number; i++) {
	if (cvtTable->papBrkTable[i] != NULL) {
	    for (j = 0; j < pTcvtTable->papBrkTable[i]->number; j++) {
		cvtTable->papBrkTable[i]->papBrkInt[j] = (void *)
			((char *)cvtTable->papBrkTable[i]->papBrkInt[j]
			-(char *)cvtTable);
	    }
	    cvtTable->papBrkTable[i]->papBrkInt = (void *)
		((char *)cvtTable->papBrkTable[i]->papBrkInt
		-(char *)cvtTable);
	    cvtTable->papBrkTable[i]->name = (void *)
		((char *)cvtTable->papBrkTable[i]->name
		-(char *)cvtTable);
	    cvtTable->papBrkTable[i] = (void *)
		((char *)cvtTable->papBrkTable[i]
		-(char *)cvtTable);
	}
    }
    cvtTable->papBrkTable = (void *)
	((char *)cvtTable->papBrkTable
	-(char *)cvtTable);
    /* write output file */
    memset((char *)&sdrHeader,0, sizeof(sdrHeader));
    sdrHeader.magic = DBMAGIC;
    sdrHeader.nbytes = Size1;
    sdrHeader.type = SDR_CVTTABLE;
    if ((cvtTable_fp = fopen(cvtSdr, "w")) == NULL) {
	printf("bldCvtTable: fopen error errno=%i\nfile=%s\n", errno, plineIn->fname);
	return (-1);
    }
    if ((rtnval = fwrite(&sdrHeader, sizeof(sdrHeader), 1,
			 cvtTable_fp)) != 1) {
	printf("bldCvtTable: fwrite error. errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, cvtSdr);
	return (-1);
    }
    if ((rtnval = fwrite(cvtTable, (int) Size1, 1, cvtTable_fp)) != 1) {
	printf("bldCvtTable: fwrite error. errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, cvtSdr);
	return (-1);
    }
    if ((fclose(cvtTable_fp)) == EOF) {
	printf("bldCvtTable: fclose error. errno=%i\nfile=%s\n",
	       errno, cvtSdr);
	return (-1);
    }
    /* allocate, init and write the contiguous area for choiceCvt.sdr */
    SSize2 = (SSize2 + sizeof(long) - 1) / sizeof(long) * sizeof(long);
    Size2 += SSize2;
    choiceCvt = dbCalloc(1,Size2);
    addr2 = (char *)choiceCvt + sizeof(struct choiceSet);
    choiceCvt->number = pTchoiceCvt->number;
    choiceCvt->papChoice = addr2;
    addr2 = (char *)addr2 + choiceCvt->number * sizeof(long);
    addrS2 = (char *)choiceCvt + Size2 - SSize2;
    for (i = 0; i < choiceCvt->number; i++) {
	if (pTchoiceCvt->papChoice[i] != NULL) {
	    choiceCvt->papChoice[i] = addrS2;
	    len = strlen(pTchoiceCvt->papChoice[i]);
	    memcpy(addrS2,pTchoiceCvt->papChoice[i],len + 1);
	    addrS2 = (char *)addrS2 + len + 1;
	}
    }
    /* make all pointers offsets */
    for (i = 0; i < choiceCvt->number; i++) {
	if (choiceCvt->papChoice[i] != NULL) {
	    choiceCvt->papChoice[i] = (void *)
		((char *)choiceCvt->papChoice[i] - (char *)choiceCvt);
	}
    }
    choiceCvt->papChoice = (void *)
	((char *)choiceCvt->papChoice - (char *)choiceCvt);
    /* write output file */
    memset((char *)&sdrHeader,0, sizeof(sdrHeader));
    sdrHeader.magic = DBMAGIC;
    sdrHeader.nbytes = Size2;
    sdrHeader.type = SDR_CHOICECVT;
    if ((choiceSdr_fp = fopen(choiceSdr, "w")) == NULL) {
	printf("bldCvtTable: fopen error errno=%i\nfile=%s\n", errno, plineIn->fname);
	return (-1);
    }
    if ((rtnval = fwrite(&sdrHeader, sizeof(sdrHeader), 1,
			 choiceSdr_fp)) != 1) {
	printf("bldCvtTable: fwrite error. errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, choiceSdr);
	return (-1);
    }
    if ((rtnval = fwrite(choiceCvt, Size2, 1,
			 choiceSdr_fp)) != 1) {
	printf("bldCvtTable: fwrite error. errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, choiceSdr);
	return (-1);
    }
    if ((fclose(choiceSdr_fp)) == EOF) {
	printf("bldCvtTable: fclose error. errno=%i\nfile=%s\n",
	       errno, choiceSdr);
	return (-1);
    }
    return (0);
}				/* end of bldCvtTable.c */
/****************************************************************************
    print message
****************************************************************************/
UnixErrMessage(mess)
    char           *mess;
{
    printf("%s\n", mess);
}
/****************************************************************************
    build the tables
****************************************************************************/
buildTables(fp, plineIn, pInfo, pbrk, pTcvtTable, pTchoiceCvt, pSize1, pSSize1, pSSize2)
    FILE           *fp;
    struct lineIn  *plineIn;
    struct inLineInfo *pInfo;
    struct brkCreateInfo *pbrk;
    struct arrBrkTable *pTcvtTable;
    struct choiceSet *pTchoiceCvt;
    long           *pSize1;
    long           *pSSize1;
    long           *pSSize2;
{
    void          *ptemp;
    int             no_fields;
    long            no_breaks;
    int             rtnval;
    int             len;
    int             slot;
    int             type;
    int             j;
    int             svinx;
    int             inx = MAX_RES_CTRL_LINES; /* start after reserved slots */
    FOREVER {
	if ((getSdrLine(fp, plineIn, TWO_QUOTES)) < 0) {
	    printf("bldCvtTable: buildTables - file %s\n", plineIn->fname);
	    return (-1);
	}
	if (plineIn->eof == EOF) {
	    return (0);
	}
	if ( (getCtrlLineType(pInfo, plineIn, pbrk, &type)) < 0 )
		return (-1);
	if ((slot = findReservedNameSlot(pInfo->brkName)) >= 0 ) {
		svinx = inx;
		inx = slot; /* force to reserved slot */
	}
	len = strlen(pInfo->brkName);
	pTchoiceCvt->papChoice[inx] = dbCalloc(1,len + 1);
	memcpy(pTchoiceCvt->papChoice[inx],pInfo->brkName, len + 1);
	*pSSize2 += len + 1;
        /* restore index if it was forced */
	if (slot >= 0 ) {
		inx = svinx;
	}
	/* Only therm tables for CTRL_TYPE2  */
	if (type == CTRL_TYPE2 ) {
	    pTcvtTable->papBrkTable[inx] = dbCalloc(1,sizeof(struct brkTable));
	    *pSize1 += sizeof(struct brkTable);
	    /* number of entries expected */
	    no_fields = (pbrk->tblEngLast - pbrk->tblEngFirst)
		/ pbrk->tblEngDelta + 1;
	    pbrk->nTable = no_fields;
	    /* allocate space for input therm data table */
	    ptemp = dbCalloc(1,pbrk->nTable * sizeof(float));
	    plineIn->pnext = NULL;
	    /* read the data */
	    if ((rtnval = read_data(fp, ptemp,
				    no_fields, pInfo->brkName, plineIn)) < 0)
		return (-1);
	    if (rtnval < no_fields) {
		printf("bldCvtTable: Error - Too few data entries in %s \n",
		       pInfo->brkName);
		printf("line=%d	%s\n", plineIn->line_no, plineIn->fname);
		return (-1);
	    }
	    pbrk->pTable = ptemp;
	    /* create a break point table */
	    if ((create_break(pbrk, brkInt, (long) MAX_BREAKS, &no_breaks)) < 0)
		return (-1);
	    pTcvtTable->papBrkTable[inx]->name = dbCalloc(1,len + 1);
	    *pSSize1 += len + 1;
	    memcpy(pTcvtTable->papBrkTable[inx]->name,pInfo->brkName,len + 1);
	    pTcvtTable->papBrkTable[inx]->number = no_breaks;
	    pTcvtTable->papBrkTable[inx]->rawLow = pbrk->rawLow;
	    pTcvtTable->papBrkTable[inx]->rawHigh = pbrk->rawHigh;
	    pTcvtTable->papBrkTable[inx]->papBrkInt =
		dbCalloc(1,no_breaks * sizeof(void *));
	    *pSize1 += no_breaks * sizeof(void *);
	    for (j = 0; j < no_breaks; j++) {
		pTcvtTable->papBrkTable[inx]->papBrkInt[j] =
			dbCalloc(1,sizeof(struct brkInt));
		*pSize1 += sizeof(struct brkInt);
		memcpy((char *)pTcvtTable->papBrkTable[inx]->papBrkInt[j],
		      (char *)&brkInt[j],
		      sizeof(struct brkInt));
	    }			/* end for */
	    free((void *) ptemp);
	/*end	 if (type == CTRL_TYPE2 ) */
	} else if (type == CTRL_TYPE3 ) {
	/* manual breakpoint tables */
	pTcvtTable->papBrkTable[inx] = dbCalloc(1,sizeof(struct brkTable));
	*pSize1 += sizeof(struct brkTable);
	/* read the data */
	if ((read_data2(fp, plineIn, pbrk, brkInt,
			 (long) MAX_BREAKS, &no_breaks)) < 0)
		return (-1);
	pTcvtTable->papBrkTable[inx]->name = dbCalloc(1,len + 1);
	*pSSize1 += len + 1;
	memcpy(pTcvtTable->papBrkTable[inx]->name,pInfo->brkName,len + 1);
	pTcvtTable->papBrkTable[inx]->number = no_breaks;
	pTcvtTable->papBrkTable[inx]->rawLow = pbrk->rawLow;
	pTcvtTable->papBrkTable[inx]->rawHigh = pbrk->rawHigh;
	pTcvtTable->papBrkTable[inx]->papBrkInt =
		dbCalloc(1,no_breaks * sizeof(void *));
	*pSize1 += no_breaks * sizeof(void *);
	
	for (j = 0; j < no_breaks; j++) {
	    pTcvtTable->papBrkTable[inx]->papBrkInt[j] =
			dbCalloc(1,sizeof(struct brkInt));
	    *pSize1 += sizeof(struct brkInt);
	    memcpy((char *)pTcvtTable->papBrkTable[inx]->papBrkInt[j],
		      (char *)&brkInt[j], sizeof(struct brkInt));
	    }			/* end for */
	} /*end	 if (type == CTRL_TYPE3 ) */
	if ( slot < 0 )
	    inx++;
    }
}
/****************************************************************************
 Create a breakpoint array from table
 return	(-1,0)	error,success
****************************************************************************/
/*
 * CREATE_BREAK
 */
static int
create_break(pbci, pabrkInt, max_breaks, n_breaks)
    struct brkCreateInfo *pbci;
    struct brkInt  *pabrkInt;
    long            max_breaks;
    long           *n_breaks;
{
    struct brkInt  *pbrkInt;
    float          *table = pbci->pTable;
    long            ntable = pbci->nTable;
    double          ilow,
                    ihigh,
                    tbllow,
                    tblhigh,
                    slope,
                    offset;
    int             ibeg,
                    iend,
                    i,
                    inc,
                    imax,
                    n;
    double          rawBeg,
                    engBeg,
                    rawEnd,
                    engEnd,
                    engCalc,
                    engActual,
                    error;
    int             valid,
                    all_ok,
                    expanding;
    /* make checks to ensure that brkCreateInfo makes sense */
    if (pbci->engLow >= pbci->engHigh) {
	UnixErrMessage("create_break: engLow >= engHigh");
	return (-1);
    }
    if ((pbci->engLow < pbci->tblEngFirst)
	    || (pbci->engHigh > pbci->tblEngLast)) {
	UnixErrMessage("create_break: engLow > engHigh");
	return (-1);
    }
    if (pbci->tblEngDelta <= 0.0) {
	UnixErrMessage("create_break: tblEngDelta <= 0.0");
	return (-1);
    }
    if (ntable < 3) {
	UnixErrMessage("raw data must have at least 3 elements");
	return (-1);
    }
/***************************************************************************
			 Convert Table to raw values
 *
 * raw and table values are assumed to be related by an equation of the form:
 *
 * raw = slope*table + offset
 *
 * The following algorithm converts each table value to raw units
 *
 * 1) Finds the locations in Table corresponding to engLow and engHigh
 *    Note that these locations need not be exact integers
 * 2) Interpolates to obtain table values corresponding to engLow and enghigh
 *    we now have the equations:
 *    rawLow = slope*tblLow + offset
 *    rawHigh = slope*tblHigh + offset
 * 4) Solving these equations for slope and offset gives:
 *    slope=(rawHigh-rawLow)/(tblHigh-tblLow)
 *    offset=rawHigh-slope*tblHigh
 * 5) for each table value set table[i]=table[i]*slope+offset
 *************************************************************************/
    /* Find engLow in Table  and then compute tblLow */
    ilow = (pbci->engLow - pbci->tblEngFirst) / (pbci->tblEngDelta);
    i = (int) ilow;
    if (i >= ntable - 1)
	i = ntable - 2;
    tbllow = table[i] + (table[i + 1] - table[i]) * (ilow - (float) i);
    /* Find engHigh in Table and then compute tblHigh */
    ihigh = (pbci->engHigh - pbci->tblEngFirst) / (pbci->tblEngDelta);
    i = (int) ihigh;
    if (i >= ntable - 1)
	i = ntable - 2;
    tblhigh = table[i] + (table[i + 1] - table[i]) * (ihigh - (float) i);
    /* compute slope and offset */
    slope = (pbci->rawHigh - pbci->rawLow) / (tblhigh - tbllow);
    offset = pbci->rawHigh - slope * tblhigh;
    /* convert table to raw units */
    for (i = 0; i < ntable; i++)
	table[i] = table[i] * slope + offset;

/*****************************************************************************
 *		Now create break point table
 *
 * The algorithm does the following:
 *
 * It finds one breakpoint interval at a time. For each it does the following:
 *
 * 1) Use a relatively large portion of the remaining table as an interval
 * 2) It attempts to use the entire interval as a breakpoint interval
 *	Success is determined by the following algorithm:
 *	  a) compute the slope using the entire interval
 *        b) for each table entry in the interval determine the eng value
 *	     using the slope just determined.
 *	  c) compare the computed value with eng value associated with table
 *        d) if all table entries are within the accuracy desired then success.
 * 3) If successful then attempt to expand the interval and try again.
 *    Note that it is expanded by up to 1/10 of the table size.
 * 4) If not successful reduce the interval by 1 and try again.
 *    Once the interval is being decreased it will never be increased again.
 * 5) The algorithm will ultimately fail or will have determined the optimum
 *    breakpoint interval
 *************************************************************************/

    /* Must start with table entry corresponding to engLow; */
    i = ilow;
    if (i >= ntable - 1)
	i = ntable - 2;
    rawBeg = table[i] + (table[i + 1] - table[i]) * (ilow - (float) i);
    engBeg = pbci->engLow;
    ibeg = (int) (ilow);       /* Make sure that ibeg > ilow */
    if( ibeg < ilow ) ibeg = ibeg + 1;
    /* start first breakpoint interval */
    n = 1;
    pbrkInt = pabrkInt;
    pbrkInt->raw = rawBeg;
    pbrkInt->eng = engBeg;
    /* determine next breakpoint interval */
    while ((engBeg <= pbci->engHigh) && (ibeg < ntable - 1)) {
	/* determine next interval to try. Up to 1/10 full range */
	iend = ibeg;
	inc = (int) ((ihigh - ilow) / 10.0);
	if (inc < 1)
	    inc = 1;
	valid = TRUE;
	/* keep trying intervals until cant do better */
	expanding = TRUE;	/* originally we are trying larger and larger
				 * intervals */
	while (valid) {
	    imax = iend + inc;
	    if (imax >= ntable) {
		/* don't go past end of table */
		imax = ntable - 1;
		inc = ntable - iend - 1;
		expanding = FALSE;
	    }
	    if (imax > (int) (ihigh + 1.0)) {	/* Don't go to far past
						 * engHigh */
		imax = (int) (ihigh + 1.0);
		inc = (int) (ihigh + 1.0) - iend;
		expanding = FALSE;
	    }
	    if (imax <= ibeg)
		break;		/* failure */
	    rawEnd = table[imax];
	    engEnd = pbci->tblEngFirst + (float) imax *(pbci->tblEngDelta);
	    slope = (engEnd - engBeg) / (rawEnd - rawBeg);
	    all_ok = TRUE;
	    for (i = ibeg + 1; i <= imax; i++) {
		engCalc = engBeg + slope * (table[i] - rawBeg);
		engActual = pbci->tblEngFirst + ((float) i) * (pbci->tblEngDelta);
		error = engCalc - engActual;
		if (error < 0.0)
		    error = -error;
		if (error >= pbci->accuracy) {
		    /* we will be trying smaller intervals */
		    expanding = FALSE;
		    /* just decrease inc and let while(valid) try again */
		    inc--;
		    all_ok = FALSE;
		    break;
		}
	    }			/* end for */
	    if (all_ok) {
		iend = imax;
		/* if not expanding we found interval */
		if (!expanding)
		    break;
		/* will automatically try larger interval */
	    }
	}			/* end while(valid) */
	/* either we failed or optimal interval has been found */
	if ((iend <= ibeg) && (iend < (int) ihigh)) {
	    UnixErrMessage("Could not meet accuracy criteria");
	    return (-1);
	}
	pbrkInt->slope = slope;
	/* get ready for next breakpoint interval */
	if (n++ >= max_breaks) {
	    UnixErrMessage("Break point table too large");
	    return (-1);
	}
	ibeg = iend;
	pbrkInt++;
	rawBeg = rawEnd;
	engBeg = engEnd;
	pbrkInt->raw = rawBeg;
	pbrkInt->eng = engBeg + (pbrkInt->raw - rawBeg) * slope;
    }
    pbrkInt->slope = 0.0;
    *n_breaks = n;
    return (0);
}
/****************************************************************************
  copies next field from file to buffer
  return	(-1,0)	error (includes EOF as error),	success
****************************************************************************/
/*
 * GETSTRING
 */
static int
getstring(fp, pbuff, plineIn)
    FILE           *fp;
    char           *pbuff;
    struct lineIn  *plineIn;
{
    char           *Wptr;
    int             no_byts;
    int             anotherfield;
    /* if line empty, try to get another */
    if (plineIn->pnext == NULL) {
	if ((getSdrLine(fp, plineIn, NO_QUOTES)) < 0)
	    return (-1);
	if (plineIn->eof == EOF) {
	    printf("bldCvtTable: Error - Unexpected EOF");
	    return (-1);
	}
    }
    /* copy field string to buff */
    sscanf(plineIn->pnext, "%s%n", pbuff, &no_byts);
    /* step to next field */
    plineIn->pnext += no_byts;
    /* test rest of line empty */
    Wptr = plineIn->pnext;
    anotherfield = 0;
    while (*Wptr != '\n') {
	if (isspace(*Wptr++)) {
	    continue;
	} else {
	    anotherfield = 1;
	    break;
	}
    }
    if (!anotherfield)
	plineIn->pnext = NULL;
    return (0);
}
/****************************************************************************
  reads data table to an array
  return	(-1,num) error , number of fields in table
****************************************************************************/
/*
 * READ_DATA
 */
static int
read_data(fp, pTempArr, nflds, pbrkName, plineIn)
    FILE           *fp;
    float          *pTempArr;
    int             nflds;
    char           *pbrkName;
    struct lineIn  *plineIn;
{
    char            buff[80];
    float           tmp;
    int             num = 0;
    FOREVER {
	if ((getstring(fp, buff, plineIn)) < 0)
	    return (-1);

	if ((strncmp(buff, "ENDTABLE", 8)) == SAME)
	    break;

	if ((sscanf(buff, "%f", &tmp)) != 1) {
	    printf("bldCvtTable: Error - Unexpected field\n");
	    printf("line=%d	%s\n", plineIn->line_no, plineIn->fname);
	    return (-1);
	}
	num++;
	if (num > nflds) {
	    printf("bldCvtTable: Error - Too many data entries in %s \n", pbrkName);
	    printf("line=%d	%s\n", plineIn->line_no, plineIn->fname);
	    return (-1);
	}
	*pTempArr++ = tmp;
    }
    return (num);
}
/****************************************************************************
Search for reserved names
	return slot number or -1 if not found. 
****************************************************************************/
/*
 * findReservedNameSlot
 */
findReservedNameSlot(pname)
    char          *pname;
{
    char            **ptr;
    int             slot;

    ptr = brkNames;
    for (slot = 0; slot < MAX_RES_CTRL_LINES ; slot++, ptr++) {
	if ((strcmp(*ptr, pname)) == SAME)
	    return (slot);
    }
    return (-1);
}
/****************************************************************************
    build index list on 1st pass
****************************************************************************/
/*
 * BLDINDEXLIST
 */
bldIndexList(fp, plineIn, pList, pInfo, pbrk, pmax_inx)
    FILE           *fp;
    struct lineIn  *plineIn;
    ELLLIST          *pList;
    struct inLineInfo *pInfo;
    struct brkCreateInfo *pbrk;
    int             *pmax_inx;
{
    static int max_inx = 0;
    int             slot;
    int             type;
    int             svinx;
    int             inx = MAX_RES_CTRL_LINES; /* start after reserved slots */
    int             res_cnt = 0;

    FOREVER {
	if ((getSdrLine(fp, plineIn, TWO_QUOTES)) < 0) {
	    printf("bldCvtTable: bldIndexList - file %s\n", plineIn->fname);
	    return (-1);
	}
	if (plineIn->eof == EOF) {
	    if ( res_cnt != MAX_RES_CTRL_LINES ) {
		printf("bldCvtTable: required entries in cvtHeader.ascii missing\n");
		return (-1);
	    }
	    *pmax_inx = max_inx;
	    return (0);		/* normal eof */
	}
	if ( (getCtrlLineType(pInfo, plineIn, pbrk, &type)) < 0 )
		return (-1);
	if ((slot = findReservedNameSlot(pInfo->brkName)) >= 0 ) {
		svinx = inx;
		inx = slot; /* force to reserved slot */
		res_cnt++;
	}
	if ((appendIndexToList(pList, pInfo, inx)) < 0)
	    return (-1);

        /* restore index if it was forced */
	if (slot >= 0 ) {
		inx = svinx;
	}
	max_inx = (max_inx < inx ) ? inx : max_inx;

	if (type == CTRL_TYPE2 || type == CTRL_TYPE3 ) {
	/* skip therm tables for CTRL_TYPE2 - skip bp tbls for CTRL_TYPE3 */
	    FOREVER { 
		if ((getSdrLine(fp, plineIn, NO_QUOTES)) < 0) {
		    printf("bldCvtTable: bldIndexList - file %s\n", plineIn->fname);
		    return (-1);
		}
		if (plineIn->eof == EOF) {
		    printf("bldCvtTable: no ENDTABLE flag for table# %d\n", inx);
		    return (-1);
		}
		if ((strncmp(plineIn->line, "ENDTABLE", 8)) == SAME)
		    break;
	    }
	}
	if ( slot < 0 )
	    inx++;
    }
}
/****************************************************************************
    Appends an index number to list
****************************************************************************/
/*
 * APPENDINDEXTOLIST
 */
static int
appendIndexToList(pList, pInfo, inx)
    ELLLIST          *pList;
    struct inLineInfo *pInfo;
    int             inx;
{
    ENODE          *node;
    int             len;
    node = dbCalloc(1,sizeof(ENODE));
    ellAdd(pList, (ELLNODE *) node);
    node->index_num = inx;
    len = strlen(pInfo->brkName);
    memcpy(node->brkName,pInfo->brkName,len + 1);
    return (0);
}
/****************************************************************************
    process control line - (0,-1) success, failure
****************************************************************************/
/*
 * GETCTRLLINETYPE
 */
getCtrlLineType(pInfo, plineIn, pbrk, ptype)
    struct inLineInfo *pInfo;
    struct lineIn  *plineIn;
    struct brkCreateInfo *pbrk;
    int            *ptype;
{
    int             num = 0;
    int             len = 0;
    int             type3found = 0;
    char            keywordString[50];

    /* scan control line */
    num = sscanf(plineIn->line, "\"%50[^\"]\" %s",
	   pInfo->brkName,
	   keywordString);
    if ( num == 2 ) {
        len = strlen (keywordString);
        if ( (strcmp ("BreakTable",keywordString)) == 0) {
	    type3found = 1;	
        }
    }
    if ( !type3found ) {
    /* re-scan control line */
        num = sscanf(plineIn->line, "\"%50[^\"]\"%f%li%f%li%f%f%f%f",
	   pInfo->brkName,
	   &pbrk->engLow,
	   &pbrk->rawLow,
	   &pbrk->engHigh,
	   &pbrk->rawHigh,
	   &pbrk->accuracy,
	   &pbrk->tblEngFirst,
	   &pbrk->tblEngLast,
	   &pbrk->tblEngDelta);
    }
    if (num == 1) {
	*ptype = CTRL_TYPE1;
    } else if (num == 2) {
	*ptype = CTRL_TYPE3;
    } else if (num == 9) {
	*ptype = CTRL_TYPE2;
    } else {
	printf("bldCvtTable: getCtrlLineType - illegal control line format\n");
	printf("bldCvtTable: line=%i\n", plineIn->line_no);
	return (-1);
    }
    return (0);
}
/****************************************************************************
    Processes command line args
****************************************************************************/
/*
 * CMD_CVTTABLE
 */
static int
cmd_CvtTable(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    char           *ptrToMem;
    struct stat    *buf;
    int             c,
                    err;
    short           aflg,
                    oflg,
                    Oflg;
    aflg = oflg = Oflg = err = 0;
    buf = dbCalloc(1,sizeof(struct stat));
    while ((c = getopt(argc, argv, "a:o:O:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile, optarg);
	    if (stat(inFile, buf) == -1) {
		printf("bldCvtTable: Can't stat file=%s errno=%i\n",
		       inFile, errno);
		err++;
	    }
	    break;
	case 'O':
	    oflg++;
	    strcpy(choiceSdr, optarg);
	    break;
	case 'o':
	    Oflg++;
	    strcpy(cvtSdr, optarg);
	    break;
	}
    if (!aflg || !oflg || !Oflg || err) {
	printf("Usage: \n");
	printf(" <bldCvtTable -a cvtTable.ascii -o cvtTable.sdr -O choiceCvt.sdr >\n");
	return (-1);
    }
    return (0);
}
/****************************************************************************
 reads data for the manual breakpoint table
 and calculates the slope for the interval
****************************************************************************/
/*
 * READ_DATA2
 */
read_data2(fp, plineIn, pbrk,  pabrkInt, max_breaks, n_breaks)
    FILE           *fp;
    struct lineIn  *plineIn;
    struct brkCreateInfo *pbrk;
    struct brkInt  *pabrkInt;
    long            max_breaks;
    long           *n_breaks;
{
    FNODE          *pNew;
    FNODE          *pfnode;
    FNODE          *pfnext;
    ELLLIST        *pElist;
    float           slope;
    int             n = 0;
    int             i = 0;
    struct brkInt  *pbrkInt;
    long            rawLow;	/* Raw value for EngLow			 */
    long            rawHigh;	/* Raw value for EngHigh		 */
    /* create a list */
    if ((pElist = (ELLLIST *) dbCalloc(1, sizeof(ELLLIST))) == NULL) {
	printf("bldCvtTable: read_data2() - dbCalloc fail\n");
	return (-1);
    }
    ellInit(pElist);

    FOREVER {
	if ((getSdrLine(fp, plineIn, NO_QUOTES)) < 0) {
	    printf("bldCvtTable: read_data2() - file %s\n", plineIn->fname);
	    return (-1);
	}
	if (plineIn->eof == EOF) {
	    printf("bldCvtTable: read_data2() Error - Unexpected EOF");
	    return (-1);
	}
	if ((strncmp(plineIn->line, "ENDTABLE", 8)) == SAME)
	    break;
	/* create a node to place the beg raw and conv eng number */
	if ((pNew = (FNODE *) dbCalloc(1, sizeof(FNODE))) == NULL) {
	    printf("bldCvtTable: read_data2() - dbCalloc fail\n");
	    return (-1);
	}
	/* reads the raw and conv eng number into node */
	if ((sscanf(plineIn->line, "%ld %f", &pNew->col1, &pNew->col2)) != 2) {
	    printf("bldCvtTable: Error - Unexpected format\n");
	    printf("line=%d	%s\n", plineIn->line_no, plineIn->fname);
	    return (-1);
	}
	/* add node to list */
	ellAdd(pElist, (ELLNODE *) pNew);
	/* get ready for next breakpoint interval */
	if (n++ >= max_breaks) {
	    UnixErrMessage("Break point table too large");
	    return (-1);
	}
    }
    if (n<2) {
	UnixErrMessage("BreakTable must have at least 2 elements");
	return (-1);
    }
    /* start first breakpoint interval */
    pbrkInt = pabrkInt;
    for (i=1, pfnode = (FNODE *) ellFirst(pElist);
	    pfnode != NULL && i < n ;
	    pfnode = (FNODE *) ellNext((ELLNODE *) pfnode), pbrkInt++, i++) {
	pfnext = (FNODE *) ellNext((ELLNODE *) pfnode);
	if ( i == 1 ) {
	    pbrk->rawLow = pfnode->col1;
	}
	slope = (pfnext->col2 - pfnode->col2) / (pfnext->col1 - pfnode->col1);
	pbrkInt->slope = slope;
	/* raw value for beginning of interval */
	pbrkInt->raw = pfnode->col1;
	/* converted value for beginning of interval */
	pbrkInt->eng = pfnode->col2;
    }
    /* last one */
    pbrkInt->raw = pfnode->col1;
    pbrkInt->eng = pfnode->col2;
    pbrkInt->slope = 0.0;
    pbrk->rawHigh = pfnode->col1;
    * n_breaks = n;
}
