/*  bldRecDef.c               $Id: bldRecDef.c,v 1.31 1995/03/21 21:47:49 jbk Exp $
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	06-07-91	rcz	initial release
 * .02	09-16-91	rcz	Complete rewrite - added automatic
 *                              pad generation and removed dimension
 *                              requirement from ascii files.
 * .03  03-26-92        rcz     from 4 to 8 byte boundary for DBF_INLINK DBF_OUTLINK DBF_FWDLINK
 * .04  05-18-92        rcz     New database access (changed sdrLoad to dbRead)
 * .05  08-03-92        rcz     Removed all references to dbr type
 * .06  08-10-92        rcz     changed a diagnostic message
 * 	...
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <dbDefs.h>
#include <dbBase.h>
#include <dbStaticLib.h>
#include <dbRecDes.h>
#include <dbFldTypes.h>
#include <dbRecType.h>
#include <sdrHeader.h>
#include "getSdrLine.h"
#include "bldLib.h"
#include <link.h>
#include <special.h>

#define BYTES20		20
#define BYTES50		50
#define RECTYP_DETECTED 1
#define ELEMENT_DETECTED 2
#define SAME 0
#define SCAN_TYPE	2
#define SCAN_COM_FLDS	3
#define SCAN_STRING	4
#define SCAN_NOACCESS	5
#define SCAN_CHOICE_SET	6
#define SCAN_INIT_VAL	7
#define SCAN_PMT_FLAG	8
#define SCAN_RANGE1	9
#define SCAN_RANGE2	10
#define SCAN_CVT_TYPE	11
/* I/O filenames */
static char     in1[BYTES50];
static char     in2[BYTES50];
static char     out1[BYTES50];
static char     out2[BYTES50];
static char     out3[BYTES50];

/* forward references */
static int      addPadToList();
static int      add_member();
static void     add_sp();
static int      appendElementToList();
static int      chkFieldAlignment();
static int      chk_align();
static int      cmd_bldRecDef();
static int      conv_strng();
static void     copyDescriptorPtrs();
static int      fieldDecode();
static int      getRangeFld();
static int      get_num_from_string();
static int      inScan();
static void     order_fields();
static void     plantRecordInfo();
static int      plantVarRangeInfo();
static int      searchForFldnum();
static void     set_dflt();
static int      verifyType();
static int      wrt_buf();
static int      wrt_com();


typedef struct var_flds {
    char            lowfl[BYTES20];	/* lowfl VAR string */
    char            highfl[BYTES20];	/* highfl VAR string */
}               VFLDS;

/* input line work area */
static struct inLineInfo {
    struct fldDes   desc;
    VFLDS           varString;
    char            prompt[MAXLINE];
    char            fldname[6];
    char            hl_Var[BYTES20];	/* holds hi or lo VAR string */
    char            initial[BYTES20];
    char            text_code[MAXLINE];	/* if field_type is dbf_noaccess */
    char            rectyp_keywd[BYTES20];
}               inLineInfo;

struct lineIn   lineIn;		/* buffer for getSdrLine */
char            ascii_name[BYTES50];	/* current record name during input */

struct sdrHeader sdrHeader;

typedef struct recList {
    SLIST         **papList;
}               SREC;

typedef struct enode {
    SNODE          *node;
    struct fldDes   desc;
    VFLDS           varString;
    char            line[MAXLINE];
    char            text_code[MAXLINE];
    char            ascii_name[BYTES50];
}               ENODE;

/* bottom part of all header files */
static char strend[] = "\n};\ntypedef struct %s %s;\n#endif\n\n";
int             errno;

/*
 *
 * BLDRECDEF.C
 *
 * MAINLINE
 *
 * INPUT:
 *       File     dbRecType.sdr
 *
 *       File     dbRecDes.ascii
 *
 *                Built from dbCommon.ascii and <record type>.ascii files
 *                which have been concatenated and processed by cpp.
 *                Note: dbCommon.ascii must be the first file in
 *                      dbRecDes.ascii.
 *
 * OUTPUT:
 *       File     dbRecDes.sdr
 *
 *       Files    <dbCommon, aiRecord, etc>.h
 *
 *	The SDR field lines from dbCommon.ascii are prepended to
 *      each of the record type description files.
 *	Data is collected in temporary lists.
 *      At the end of each record several sorted arrays are created.
 *      When all records have been processed, an appropriate sdrHeader
 *      is placed at the beginning of the sdr file. The data structures
 *      are written. Note: any pointers are converted to an offset
 *      before they are written.
 *
 *      Appropriate header files are built for dbCommon and each
 *      record type
 */

struct recType *dbRecType=NULL;
struct recDes  *dbRecDes=NULL;
struct dbBase  *pdbBase=NULL;

main(argc, argv)
    int             argc;
    char          **argv;
{
    FILE           *in1_fp,
                   *fopen();
    FILE           *out1_fp;
    FILE           *out2_fp;
    FILE           *out3_fp;

    SREC           *pRecTypeList = NULL;
    struct lineIn  *plineIn = &lineIn;	/* ptr to buffer */
    struct inLineInfo *pInfo = &inLineInfo;	/* input line work area */
    struct recTypDes *pRecTypDes = NULL;
    struct recDes  *pdbRecDes = NULL;
    struct recDes  *wptr = NULL;
    SNODE          *pList = NULL;
    SNODE          *pt = NULL;
    ENODE          *svnode = NULL;
    void           *ptrToRecTypDes = NULL;
    void           *ptrToPtrAry = NULL;
    char            buffer[MAXBUF];
    char            sNameBuf[BYTES50];
    long            rsize = 0;
    long            size = 0;
    long            baddr = 0;
    long            baddr2 = 0;
    long            base1 = 0;
    int             delta = 0;
    int             rec_no = 0;
    int             status = 0;
    int             i = 0;
    int             j = 0;
    int             pad_no = 0;
    int             CurOffSet = 0;
    int             no_fields;
    int             firstRecTypeFlag;
    int             dbCommonFlag;
    int             firstElement;
    int             dbCommonOffSet = 0;
    int             pad_size;
    int             element_count;
    int             record_element_count;
    int             no_prompt = 0;
    int             no_links = 0;
    int             last_offset = 0;
    FILE	    *fp;
    int		    flddespad;
    if (chk_align() < 0)
	return (-1);
    if ((cmd_bldRecDef(argc, argv)) < 0) {
	return (-1);
    }
    plineIn->fname = in1;
    /* method to get dbRead to skip the checksum */
    pdbBase = dbAllocBase();
    pdbBase->pvtSumFlag = 1;
    /* set default record name to "ai" */
    strcpy(ascii_name,"ai");

    if ((in1_fp = fopen( in1, "r")) == NULL) {
	printf("bldRecDef: fopen error errno=%i\nfile=%s\n", errno, in1);
	return (-1);
    }
    /* load ascii record type definitions - dbRecType.sdr */
    fp = fopen(in2,"r");
    if(!fp) return(-1);
    if(dbRead(pdbBase,fp)) return(-1);
    fclose(fp);
    dbRecDes = pdbBase->precDes;
    dbRecType = pdbBase->precType;


    pRecTypeList = dbCalloc(1,sizeof(SREC));
    pRecTypeList->papList = dbCalloc(1,dbRecType->number * sizeof(void *));
    firstRecTypeFlag = TRUE;
    FOREVER {
	/* get next SDR line */
        if ((getSdrLine(in1_fp, plineIn, TWO_QUOTES)) < 0) {
            printf("\nbldRecDef: getSdrLine failed\n");
            printf("\nfile %s\n line number %d\nRecType=%s\n%s\n",
                 plineIn->fname,
                 plineIn->line_no,
                 ascii_name,
                 plineIn->line);
	    return (-1);
        }
	if (plineIn->eof == EOF)
	    break;
	memset((char *)&inLineInfo,0,sizeof(struct inLineInfo));

	/* decode the first 7 common fields of sdr line */
	if ((status = inScan(SCAN_TYPE, plineIn, pInfo)) < 0)
	    return (-1);

	switch (status) {

	case RECTYP_DETECTED:

	    /* accept dbCommon as a rectype */
	    if ((strcmp("dbCommon", ascii_name)) == SAME) {
		rec_no = 0;
		dbCommonFlag = TRUE;
	    } else
		dbCommonFlag = FALSE;

	    /* set rec_no if valid rectype */
	    if (!dbCommonFlag && (!verifyType(dbRecType, &rec_no))) {
		return (-1);
	    }
	    /* dbCommon must be 1st rectype in file */
	    if (!dbCommonFlag && firstRecTypeFlag) {
		printf("bldRecDef: dbCommon.ascii must be the first file in %s\n", in1);
		return (-1);
	    } else
		firstRecTypeFlag = FALSE;

	    if (pRecTypeList->papList[rec_no]) {
		printf("bldRecDef: RECTYPE %s is a duplicate\n", ascii_name);
		return (-1);
	    }
	    /* allocate a rectype list */
	    pRecTypeList->papList[rec_no] = dbCalloc(1,sizeof(SLIST));
	    firstElement = TRUE;

	    break;
	case ELEMENT_DETECTED:

	    if (!dbCommonFlag && firstElement)
		CurOffSet = dbCommonOffSet;
	    if (firstElement)
		firstElement = FALSE;

	    /* decode the remaining fields */
	    if ((status = fieldDecode(plineIn, pInfo)) < 0)
		return (-1);

	    /* see if an alignment pad is needed */
	    /* return	(0,-1,+N)	OK,FAIL,pad size */
	    if ((pad_size = chkFieldAlignment(plineIn, pInfo, CurOffSet)) < 0)
		return (-1);

	    /* insert a pad before the current element */
	    if (pad_size)
		if ((addPadToList(pRecTypeList->papList[rec_no], &CurOffSet, pad_size, &pad_no)) < 0)
		    return (-1);

	    /* allocate current element to the end of the current list */
	    if ((appendElementToList(pRecTypeList->papList[rec_no], plineIn, pInfo, &CurOffSet)) < 0)
		return (-1);

	    if (dbCommonFlag) {
		dbCommonOffSet = CurOffSet;
	    }
	    break;
	default:
	    break;
	}
    } /* end of input */
    if ((fclose(in1_fp)) == EOF) {
	printf("bldRecDef: fclose error. errno=%i\nfile=%s\n",
	       errno, in1);
	return (-1);
    }
    /* create the dbCommon header file */
    strcpy(out2, "dbCommon.h");
    if ((out2_fp = fopen(out2, "w")) == NULL) {
	printf("bldRecDef: fopen error errno=%i\nfile=%s\n",
	       errno, out2);
	return (-1);
    }
    strcpy(sNameBuf, "dbCommon");
    if ((wrt_com(out2_fp, buffer, out2, sNameBuf)) < 0)
	return (-1);
    pList = (SNODE *) pRecTypeList->papList[0];	/* dbCommon List */
    pt = sllFirst(pList);
    while (pt) {
	svnode = (ENODE *) pt;
	if ((add_member(out2_fp,
			buffer,
			out2,
			svnode)) < 0)
	    return (-1);
	pt = sllNext(pt);
    }
    sprintf(buffer,strend,sNameBuf,sNameBuf);
    /* strcpy(buffer, strend); */
    if ((wrt_buf(out2_fp, buffer, out2)) < 0)
	return (-1);
    if ((fclose(out2_fp)) == EOF) {
	printf("bldRecDef: fclose error. errno=%i\nfile=%s\n",
	       errno, out1);
	return (-1);
    }
    /* create the xxRecord.h header files */
    for (rec_no = 1; rec_no < dbRecType->number; rec_no++) {
	if (!pRecTypeList->papList[rec_no]) {
	    printf("bldRecDef: No List for record %d\n", rec_no);
	    return (-1);
	}
	/* open current xxRecord header file for writing */
	strcpy(out3, dbRecType->papName[rec_no]);
	strcat(out3, "Record.h");
	if ((out3_fp = fopen(out3, "w")) == NULL) {
	    printf("bldRecDef: fopen error errno=%i\nfile=%s\n",
		   errno, out3);
	    return (-1);
	}
	strcpy(sNameBuf, dbRecType->papName[rec_no]);
	strcat(sNameBuf, "Record");
	if ((wrt_com(out3_fp, buffer, out3, sNameBuf)) < 0)
	    return (-1);
	pList = (SNODE *) pRecTypeList->papList[0];	/* dbCommon List */
	pt = sllFirst(pList);
	while (pt) {
	    svnode = (ENODE *) pt;
	    if ((add_member(out3_fp,
			    buffer,
			    out3,
			    svnode)) < 0)
		return (-1);
	    pt = sllNext(pt);
	}
	/* insert  comment before record specific fields */
	strcpy(buffer, "\n\t/* start of ");
	strcat(buffer, dbRecType->papName[rec_no]);
	strcat(buffer, " specific fields */ ");
	if ((wrt_buf(out3_fp, buffer, out3)) < 0)
	    return (-1);
	pList = (SNODE *) pRecTypeList->papList[rec_no];	/* RecordN List */
	pt = sllFirst(pList);
	while (pt) {
	    svnode = (ENODE *) pt;
	    if ((add_member(out3_fp,
			    buffer,
			    out3,
			    svnode)) < 0)
		return (-1);
	    pt = sllNext(pt);
	}
	/* write the end part of the header file */
    	sprintf(buffer,strend,sNameBuf,sNameBuf);
	/* strcpy(buffer, strend); */
	if ((wrt_buf(out3_fp, buffer, out3)) < 0)
	    return (-1);
	if ((fclose(out3_fp)) == EOF) {
	    printf("bldRecDef: fclose error. errno=%i\nfile=%s\n",
		   errno, out1);
	    return (-1);
	}
    }

    /* build each record type including dbCommon in collection area */
    pdbRecDes = dbCalloc(1,sizeof(struct recDes));
    rsize += sizeof(struct recDes);
    pdbRecDes->number = dbRecType->number;
    /* record type description  ptr array */
    pdbRecDes->papRecTypDes = dbCalloc(1,dbRecType->number * sizeof(void *));
    rsize += (dbRecType->number * sizeof(void *));

    /* prepend dbCommon to each rec_no in collection area */
    /* Note: record type output slot 0 must be NULL for DCT */
    /* the first record type is numbered 1 */
    for (rec_no = 1; rec_no < dbRecType->number; rec_no++) {

	/* ptrs to rectype lists */
	if (!pRecTypeList->papList[rec_no])
	    continue;
	pdbRecDes->papRecTypDes[rec_no] = dbCalloc(1,sizeof(struct recTypDes));
	rsize += sizeof(struct recTypDes);
	/* ensure some element lines in file */
	if ((element_count = pRecTypeList->papList[0]->count) == 0) {
	    printf("bldRecDef: dbCommon.ascii has zero elements\n");
	    return (-1);
	}
	if ((record_element_count = pRecTypeList->papList[rec_no]->count) == 0) {
	    printf("bldRecDef: %sRecord.ascii has zero elements\n",
		   dbRecType->papName[rec_no]);
	    return (-1);
	}
	element_count += record_element_count;
	pdbRecDes->papRecTypDes[rec_no]->papFldDes = 
	   dbCalloc(1,(element_count * sizeof(void *)));
	rsize += (element_count * sizeof(void *));
	/*pad to double*/
	rsize += ((rsize + sizeof(double) - 1)
		/ sizeof(double) * sizeof(double)) - rsize;
	rsize += (element_count * sizeof(struct fldDes));
	no_fields = 0;
	no_links = 0;
	no_prompt = 0;
	/* calc no_fields , no_links, no_prompt, rec_size */
	plantRecordInfo(pRecTypeList->papList[0],
			pRecTypeList->papList[rec_no],
			&no_fields, &no_links, &no_prompt, &last_offset);
	pdbRecDes->papRecTypDes[rec_no]->no_fields = no_fields;
	pdbRecDes->papRecTypDes[rec_no]->no_links = no_links;
	pdbRecDes->papRecTypDes[rec_no]->no_prompt = no_prompt;
	/* round up rec_size for sparc */
	if (delta = (last_offset % 8))
	    last_offset += (8 - delta);
	pdbRecDes->papRecTypDes[rec_no]->rec_size = last_offset;

	/* process range1 and range2 info */
	if ((plantVarRangeInfo(pRecTypeList->papList[0],
			       pRecTypeList->papList[rec_no],
			       element_count)) < 0)
	    return (-1);
	/* copy the record descriptor pointers to the collection area */
	copyDescriptorPtrs(pRecTypeList->papList[0],
			   pRecTypeList->papList[rec_no],
			   &pdbRecDes->papRecTypDes[rec_no]->papFldDes[0]);
	size = no_links * sizeof(short);
	/* make sure we have padding to long words */
	size = (size + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	pdbRecDes->papRecTypDes[rec_no]->link_ind = dbCalloc(1,size);
	rsize += no_links * sizeof(short);
	rsize = (rsize + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	i = 0;
	for (j = 0; j < no_fields; j++) {
	    switch (pdbRecDes->papRecTypDes[rec_no]->papFldDes[j]->field_type) {
	    case DBF_INLINK:
	    case DBF_OUTLINK:
	    case DBF_FWDLINK:
		pdbRecDes->papRecTypDes[rec_no]->link_ind[i] = (short) j;
		i++;
		break;
	    default:
		break;
	    }
	}
	/* install sorted fields and corresponding offsets */
	pdbRecDes->papRecTypDes[rec_no]->sortFldName =
		dbCalloc(1,(no_fields * sizeof(long)));
	rsize += no_fields * sizeof(long);
	size = no_fields * sizeof(short);
	/* make sure we have padding to long words */
	size = (size + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	pdbRecDes->papRecTypDes[rec_no]->sortFldInd = dbCalloc(1,size);
	rsize += no_fields * sizeof(short);
	rsize = (rsize + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	order_fields(pdbRecDes->papRecTypDes[rec_no]);
    }
    /* init the sdrHeader header */
    memset((char *)&sdrHeader,0, sizeof(struct sdrHeader));
    sdrHeader.magic = DBMAGIC;
    sdrHeader.nbytes = rsize;
    sdrHeader.type = SDR_DB_RECDES;
    /* write output file */
    if ((out1_fp = fopen(out1, "w")) == NULL) {
	printf("bldRecDef: fopen errno=%i \nfile=%s\n",
	       errno, out1);
	return (-1);
    }
    if ((fwrite(&sdrHeader, sizeof(sdrHeader), 1,
		out1_fp)) != 1) {
	printf("bldRecDef: fwrite error. errno=%i \nfile=%s\n",
	       errno, out1);
	return (-1);
    }
/****************************************************************/
    /* record type */
/****************************************************************/
    wptr = dbCalloc(1,sizeof(struct recDes));
    /* copy structure to working area */
    memcpy((char *)wptr,(char *)pdbRecDes,sizeof(struct recDes));
    /* change ptr to offset */
    baddr = (long) wptr + (long) sizeof(struct recDes);
    wptr->papRecTypDes = (void *)((long) baddr - (long) wptr);
    /* write the data */
    if ((fwrite(wptr, (int) sizeof(struct recDes), 1, out1_fp)) != 1) {
	printf("bldRecDef: fwrite error. errno=%i \nfile=%s\n",
	       errno, out1);
	return (-1);
    }
    /* init base1 address to just after the ptr array */
    base1 = (long) wptr + (long) sizeof(struct recDes);
    base1 += (long) dbRecType->number * (long) sizeof(long);
/****************************************************************/
    /* record type description  ptr array */
/****************************************************************/
    wptr->papRecTypDes = dbCalloc(1,dbRecType->number * sizeof(void *));
    /* set baddr to just after the ptr array */
    baddr = base1;
    for (i = 0; i < dbRecType->number; i++) {
	pRecTypDes = pdbRecDes->papRecTypDes[i];
	if (pRecTypDes) {
	    /* insert ptr for the recTypDes structure */
	    /* insert and change ptr to offset */
	    wptr->papRecTypDes[i] = (void *)((long) baddr - (long) wptr);
	    baddr += (long) sizeof(struct recTypDes);
	    /* set baddr to just after the ptr array */
	    baddr += (long) pRecTypDes->no_fields * (long) sizeof(long);
	    flddespad = ((baddr + sizeof(double) - 1)
		/ sizeof(double) * sizeof(double)) - baddr;
	    baddr += flddespad;
	    baddr += pRecTypDes->no_fields * sizeof(struct fldDes);
	    baddr += pRecTypDes->no_fields * sizeof(long);
	    baddr += pRecTypDes->no_fields * sizeof(short);
	    /* make sure we have padding to long words */
	    baddr = (baddr + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	    baddr += pRecTypDes->no_links * sizeof(short);
	    /* make sure we have padding to long words */
	    baddr = (baddr + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	}
    }
    /* write the data */
    if ((fwrite(wptr->papRecTypDes, (int) dbRecType->number * sizeof(void *),
		1, out1_fp)) != 1) {
	printf("bldRecDef: fwrite error. errno=%i \nfile=%s\n",
	       errno, out1);
	return (-1);
    }
/****************************************************************/
    /* record type description - struct recTypDes */
/****************************************************************/
    baddr = base1;
    for (rec_no = 1; rec_no < dbRecType->number; rec_no++) {
	pRecTypDes = pdbRecDes->papRecTypDes[rec_no];
	if (!(pRecTypDes = pdbRecDes->papRecTypDes[rec_no])) {
	    printf("bldRecDef: Internal error - can't find rec_no=%d\n", rec_no);
	    return (-1);
	}
	ptrToRecTypDes = dbCalloc(1,sizeof(struct recTypDes));
	wptr->papRecTypDes[rec_no] = ptrToRecTypDes;
	memcpy((char *)ptrToRecTypDes,(char *)pdbRecDes->papRecTypDes[rec_no],
	      sizeof(struct recTypDes));
	baddr += (long) sizeof(struct recTypDes);
	wptr->papRecTypDes[rec_no]->papFldDes = (void *)(baddr - (long) wptr);
	/* set baddr to just after the ptr array */
	baddr += (long) pRecTypDes->no_fields * (long) sizeof(long);
	/*make 1st fldDes fall on double boundary*/
	flddespad = ((baddr + sizeof(double) - 1)
		/ sizeof(double) * sizeof(double)) - baddr;
	baddr += flddespad;
	baddr2 = baddr;

	/* set baddr to just after the last field descriptor */
	baddr += pRecTypDes->no_fields * sizeof(struct fldDes);

	/* sortFldName array */
	wptr->papRecTypDes[rec_no]->sortFldName = (void *)(baddr - (long) wptr);
	/* set baddr to just after sortFldName array */
	baddr += pRecTypDes->no_fields * sizeof(long);

	/* sortFldInd array */
	wptr->papRecTypDes[rec_no]->sortFldInd = (void *)(baddr - (long) wptr);
	baddr += pRecTypDes->no_fields * sizeof(short);

	/* make sure we have padding to long words */
	baddr = (baddr + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	/* link_ind  array */
	wptr->papRecTypDes[rec_no]->link_ind = (void *)(baddr - (long) wptr);
	baddr += pRecTypDes->no_links * sizeof(short);

	/* make sure we have padding to long words */
	baddr = (baddr + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	/* write the data */
	if ((fwrite(wptr->papRecTypDes[rec_no], (int) sizeof(struct recTypDes),
		    1, out1_fp)) != 1) {
	    printf("bldRecDef: fwrite error. errno=%i \nfile=%s\n", errno, out1);
	    return (-1);
	}
/****************************************************************/
	/* field descriptors ptr array */
/****************************************************************/
	ptrToPtrAry = dbCalloc(1,pRecTypDes->no_fields * sizeof(long));
	wptr->papRecTypDes[rec_no]->papFldDes = ptrToPtrAry;
	/* baddr2 is set to just after the field desc  ptr array */
	for (j = 0; j < pRecTypDes->no_fields; j++) {
	    /* insert each pointer to field desc */
	    wptr->papRecTypDes[rec_no]->papFldDes[j] = (void *)(baddr2 - (long) wptr);
	    baddr2 += sizeof(struct fldDes);
	}
	/* baddr2 is just after the last field descriptor */
	/* set baddr2 to just after sortFldName array */
	baddr2 += pRecTypDes->no_fields * sizeof(long);
	/* set baddr2 to just after sortFldInd array */
	baddr2 += pRecTypDes->no_fields * sizeof(short);
	/* make sure we have padding to long words */
	baddr2 = (baddr2 + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	/* link_ind  array */
	/* set baddr2 to just after link_ind array */
	baddr2 += pRecTypDes->no_links * sizeof(short);
	/* make sure we have padding to long words */
	baddr2 = ( baddr2 + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	/* write the data */
	if ((fwrite(wptr->papRecTypDes[rec_no]->papFldDes,
		 (int) wptr->papRecTypDes[rec_no]->no_fields * sizeof(long),
		    1, out1_fp)) != 1) {
	    printf("bldRecDef: fwrite error. errno=%i \nfile=%s\n",
		   errno, out1);
	    return (-1);
	}
/****************************************************************/
	/* field descriptors */
/****************************************************************/
	/*If necessary write pad*/
	if(flddespad) {
	    long temp=0;

	    if ((fwrite(&temp,flddespad,1, out1_fp)) != 1) {
		printf("bldRecDef: fwrite error. errno=%i \nfile=%s\n",
			errno,out1);
		return (-1);
	    }
	}
	for (j = 0; j < pRecTypDes->no_fields; j++) {
	    /* write each field desc directly */
	    if ((fwrite(pRecTypDes->papFldDes[j],
			(int) sizeof(struct fldDes),
			1, out1_fp)) != 1) {
		printf("bldRecDef: fwrite error. errno=%i \nfile=%s\n",
		       errno, out1);
		return (-1);
	    }
	}
/****************************************************************/
	/* sort and link arrays */
/****************************************************************/
	size = pRecTypDes->no_fields * sizeof(long);
	/* output sortFldName array directly */
	if ((fwrite(pRecTypDes->sortFldName,
		    (int) size, 1, out1_fp)) != 1) {
	    printf("bldRecDef: fwrite error. errno=%i \nfile=%s\n",
		   errno, out1);
	    return (-1);
	}
	/* output sortFldInd array directly */
	size = pRecTypDes->no_fields * sizeof(short);
	/* make sure we have padding to long words */
	size = (size + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	if ((fwrite(pRecTypDes->sortFldInd,
		    (int) size, 1, out1_fp)) != 1) {
	    printf("bldRecDef: fwrite error. errno=%i \nfile=%s\n", errno, out1);
	    return (-1);
	}
	/* output link_ind array directly */
	size = pRecTypDes->no_links * sizeof(short);
	/* make sure we have padding to long words */
	size = (size + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	if ((fwrite(pRecTypDes->link_ind,
		    (int) size, 1, out1_fp)) != 1) {
	    printf("bldRecDef: fwrite error. errno=%i \nfile=%s\n", errno, out1);
	    return (-1);
	}
	/* Note: the reverse order of these free's */
	free(ptrToPtrAry);	/* last-1 */
	free(ptrToRecTypDes);	/* last */
    }
    if ((fclose(out1_fp)) == EOF) {
	printf("bldRecDef: fclose error. errno=%i\nfile=%s\n", errno, out1);
	return (-1);
    }
    return (0);
}				/* end of main bldRecDef.c */

struct inLineInfo infoPad;	/* work area */
struct lineIn   Padline;
/****************************************************************************
   Append a pad field to the end of the list
****************************************************************************/
/*
 * ADDPADTOLIST
 */
static int
addPadToList(pList, pCurOffSet, pad_size, pPad_no)
    SLIST          *pList;
    int            *pCurOffSet;
    int             pad_size;
    int            *pPad_no;
{
    char            padPrompt[PROMPT_SZ];
    struct lineIn  *pline = &Padline;	/* pad info line */
    struct inLineInfo *pinfoPad = &infoPad;	/* ptr to Pad work area */
    struct fldDes  *pfield = (struct fldDes *) & infoPad.desc;
    *pPad_no += 1;
    memset((char *)&infoPad,0, sizeof(struct inLineInfo));
    memset(padPrompt,0, PROMPT_SZ);

    /* now fill in the Pad specifics */
    strcpy(padPrompt, "Created Pad ");
    sprintf(pinfoPad->text_code, "char		p%-03d[%d];", *pPad_no, pad_size);
    /* the following '->line' stuff is just for debugging */
    strcpy(pline->line, "INSERTED PAD - ");
    strcat(pline->line, pinfoPad->text_code);
    strcat(pline->line, "\n");

    pfield->promptflag = 0;
    memcpy(pfield->prompt,padPrompt, PROMPT_SZ - 1);
    sprintf(&pfield->fldname[0], "P%-03d", *pPad_no);
    pfield->special = SPC_NOMOD;
    pfield->field_type = DBF_NOACCESS;
    pfield->interest = 5;
    pfield->process_passive = NO;
    pfield->size = pad_size;

    /* add to the end of the current list */
    if ((appendElementToList(pList, pline, pinfoPad, pCurOffSet)) < 0)
	return (-1);
    return (0);
}

/****************************************************************************
  Add a member to the header file
****************************************************************************/
/*
 * ADD_MEMBER
 */
static int
add_member(fp, pbuf, pfilename, pnode)
    FILE           *fp;
    char           *pbuf;
    char           *pfilename;
    ENODE          *pnode;
{

    struct fldDes  *pfield = (struct fldDes *) & pnode->desc;
    char            comment[BYTES50];
    char           *pcomment = comment;
    char            fldname[BYTES50];
    char           *pfldname = fldname;
    short           i;
    /* change field name to lower case */
    i = 0;
    while (*(pfield->fldname + i) && *(pfield->fldname + i) != ' '
	    && i < FLDNAME_SZ) {
	if (isupper(*(pfield->fldname + i))) {
	    *(pfldname + i) = tolower(*(pfield->fldname + i));
	} else {
	    *(pfldname + i) = *(pfield->fldname + i);
	}
	i++;
    }
    *(pfldname + i) = NULL;
    strcpy(pcomment, pfield->prompt);

    switch (pfield->field_type) {
    case DBF_NOACCESS:
	strcpy(pbuf, "\n	");
	strcat(pbuf, pnode->text_code);
	strcat(pbuf, "		/* ");
	strcat(pbuf, pcomment);
	strcat(pbuf, " */");
	break;
    case DBF_STRING:
	sprintf(pbuf, "\n	char		%s[%d];		/* %s */",
		       pfldname, pfield->size, pcomment);
	break;
    case DBF_ENUM:
    case DBF_CVTCHOICE:
    case DBF_DEVCHOICE:
    case DBF_GBLCHOICE:
    case DBF_RECCHOICE:
	sprintf(pbuf, "\n	unsigned short	%s;		/* %s */",
		       pfldname, pcomment);
	break;
    case DBF_SHORT:
	sprintf(pbuf, "\n	short		%s;		/* %s */",
		       pfldname, pcomment);
	break;
    case DBF_USHORT:
	sprintf(pbuf, "\n	unsigned short	%s;		/* %s */",
		       pfldname, pcomment);
	break;
    case DBF_CHAR:
	sprintf(pbuf, "\n	char		%s;		/* %s */",
		       pfldname, pcomment);
	break;
    case DBF_UCHAR:
	sprintf(pbuf, "\n	unsigned char	%s;		/* %s */",
		       pfldname, pcomment);
	break;
    case DBF_LONG:
	sprintf(pbuf, "\n	long		%s;		/* %s */",
		       pfldname, pcomment);
	break;
    case DBF_ULONG:
	sprintf(pbuf, "\n	unsigned long	%s;		/* %s */",
		       pfldname, pcomment);
	break;
    case DBF_FLOAT:
	sprintf(pbuf, "\n	float		%s;		/* %s */",
		       pfldname, pcomment);
	break;
    case DBF_DOUBLE:
	sprintf(pbuf, "\n	double		%s;		/* %s */",
		       pfldname, pcomment);
	break;
    case DBF_INLINK:
    case DBF_OUTLINK:
    case DBF_FWDLINK:
	sprintf(pbuf, "\n	struct link	%s;		/* %s */",
		       pfldname, pcomment);
	break;
    default:
	printf("bldRecDef: add_member (default reached)\n");
	printf("%s - RecType=%s\n", pnode->line, pnode->ascii_name);
	return (-1);
	break;
    }
    if (wrt_buf(fp, pbuf, pfilename) < 0)
	return (-1);
    return (0);
}

/****************************************************************************
     Replaces NULLS with spaces in strings
****************************************************************************/
/*
 * ADD_SP
 */
static void
add_sp(pin, len)
    register char  *pin;
    register int    len;
{
    register int    i;
    for (i = 0; i < len; i++)
	if (*(pin + i) == 0)
	    *(pin + i) = ' ';
}

/****************************************************************************
    Appends an element to list
****************************************************************************/
/*
 * APPENDELEMENTTOLIST
 */
static int
appendElementToList(pList, pline, pInfo, pCurOffSet)
    SLIST          *pList;
    struct lineIn  *pline;
    struct inLineInfo *pInfo;
    int            *pCurOffSet;
{
    SNODE          *pNew;
    ENODE          *node;
    int             len;
    pInfo->desc.offset = *pCurOffSet;
    *pCurOffSet += pInfo->desc.size;
    pNew = dbCalloc(1,sizeof(ENODE));
    sllAdd(pList, pNew);
    node = (ENODE *) pNew;
    len = strlen(pline->line);
    memcpy(node->line,pline->line,len + 1);
    memcpy((char *)&node->desc,(char *)&pInfo->desc, sizeof(struct fldDes));
    memcpy((char *)&node->varString,(char *)&pInfo->varString,sizeof(VFLDS));
    strcpy(node->ascii_name, ascii_name);
    len = strlen(pInfo->text_code);
    memcpy(node->text_code,pInfo->text_code,len);
    return (0);
}

/****************************************************************************
    Check if pad element needed
****************************************************************************/
/*
 * CHKFIELDALIGNMENT
 * return	(0,-1,+N)	no pad needed, FAIL, pad size
 */
static int
chkFieldAlignment(plineIn, pInfo, CurOffSet)
    struct lineIn  *plineIn;
    struct inLineInfo *pInfo;
    int             CurOffSet;
{
    struct fldDes  *pfield = (struct fldDes *) & pInfo->desc;
    int             pad_size;
    switch (pfield->field_type) {
	/* don't care */
    case DBF_STRING:
    case DBF_CHAR:
    case DBF_UCHAR:
	return (0);

    case DBF_SHORT:
    case DBF_USHORT:
    case DBF_ENUM:
    case DBF_GBLCHOICE:
    case DBF_CVTCHOICE:
    case DBF_RECCHOICE:
    case DBF_DEVCHOICE:
	if (pad_size = CurOffSet % 2)
	    return (2 - pad_size);
	return (0);

    case DBF_LONG:
    case DBF_ULONG:
    case DBF_FLOAT:

	if (pad_size = CurOffSet % 4)
	    return (4 - pad_size);
	return (0);

    case DBF_DOUBLE:
    case DBF_INLINK:
    case DBF_OUTLINK:
    case DBF_FWDLINK:
	if (pad_size = CurOffSet % 8)
	    return (8 - pad_size);
	return (0);

    case DBF_NOACCESS:
	/* assume noAccess  definitions >= 8 must be mod 8 */
	if (pfield->size >= 8) {
	    if (pad_size = CurOffSet % 8)
		return (8 - pad_size);
	/* assume noAccess  definitions >= 4 must be mod 4 */
	} else if (pfield->size >= 4) {
	    if (pad_size = CurOffSet % 4)
		return (4 - pad_size);
	} else if (pfield->size == 0) {
	    break;
	/* assume noAccess  definitions == 2 must be mod 2 */
	} else if (pfield->size >= 2) {
	    if (pad_size = CurOffSet % 2)
		return (2 - pad_size);
	}
	return (0);
    default:
	printf("bldRecDef: chkFieldAlignment - default reached\n");
	break;
    }
    printf("bldRecDef: chkFieldAlignment - field alignment fail\n");
    printf("file %s line number %d\n%s - RecType=%s current CurOffSet=%d\n",
	   plineIn->fname, plineIn->line_no, plineIn->line,
	   ascii_name, CurOffSet);
    return (-1);
}

/****************************************************************************
    Checks alignment of header structs
****************************************************************************/
/*
 * CHK_ALIGN
 */
static int
chk_align()
{
    if ((sizeof(struct fldDes) % 8)) {
	printf("bldRecDef: struct %s size alignment error\n", "fldDes");
	return (-1);
    }
    if ((sizeof(struct recTypDes) % 8)) {
	printf("bldRecDef: struct %s size alignment error\n", "recTypDes");
	return (-1);
    }
    if ((sizeof(struct recDes) % 8)) {
	printf("bldRecDef: struct %s size alignment error\n", "recDes");
	return (-1);
    }
    return (0);
}

/****************************************************************************
    Processes command line args
****************************************************************************/
/*
 * CMD_BLDRECDEF
 */
static int
cmd_bldRecDef(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    struct stat    *buf;
    int             c,
                    err;
    short           aflg,
                    bflg,
                    oflg;
    aflg = bflg = oflg = err = 0;
    buf = dbCalloc(1,sizeof(struct stat));
    while ((c = getopt(argc, argv, "a:b:o:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(in1, optarg);
	    if (stat(in1, buf) == -1) {
		printf("bldRecDef: Can't stat file=%s errno=%i\n", in1, errno);
		err++;
	    }
	    break;
	case 'b':
	    bflg++;
	    strcpy(in2, optarg);
	    break;
	case 'o':
	    oflg++;
	    strcpy(out1, optarg);
	    break;
	}
    if (!bflg || !aflg || !oflg || err) {
	printf("Usage:\n");
	printf(" bldRecDef  -a dbRecDes.ascii -b dbRecType.sdr -o dbRecDes.sdr\n");
	return (-1);
    }
    return (0);
}

/****************************************************************************
    Converts string to value
****************************************************************************/
/*
 * CONV_STRNG
 */
static int
conv_strng(plineIn, pInfo, pstring, pvalue)
    struct lineIn  *plineIn;
    struct inLineInfo *pInfo;
    char           *pstring;	/* pointer to string */
    union fld_types *pvalue;	/* ptr to place to put number */
{
    double          number;
    int             err = 0;
    union fld_types tmpnum;
    if ((get_num_from_string(pstring, &number, plineIn)) < 0)
	return (-1);

    switch (pInfo->desc.field_type) {
    case (DBF_DOUBLE):
	*(double *) pvalue = (double) number;
	break;
    case (DBF_FLOAT):
	*(float *) pvalue = (float) number;
	break;
    case (DBF_CHAR):
	*(char *) pvalue = (char) number;
	if (number > 127 || number < -128)
	    err = 1;
	break;
    case (DBF_UCHAR):
	*(unsigned char *) pvalue = (unsigned char) number;
	if (number < 0 || number > 255)
	    err = 1;
	break;
    case (DBF_ENUM):
	*(unsigned short *) pvalue = (unsigned short) number;
	if (number < 0 || number > 65535)
	    err = 1;
	break;
    case (DBF_SHORT):
	*(short *) pvalue = (short) number;
	if (number > 32767 || number < -32768)
	    err = 1;
	break;
    case (DBF_USHORT):
	*(unsigned short *) pvalue = (unsigned short) number;
	if (number < 0 || number > 65535)
	    err = 1;
	break;
    case (DBF_ULONG):
	tmpnum.ulong_value = (unsigned long) number;
	*(unsigned long *) pvalue = tmpnum.ulong_value;
	break;
    case (DBF_LONG):
	*(long *) pvalue = (long) number;
	break;
    case (DBF_GBLCHOICE):
    case (DBF_CVTCHOICE):
    case (DBF_RECCHOICE):
    case (DBF_DEVCHOICE):
	*(unsigned short *) pvalue = (unsigned short) number;
	if (number < 0 || number > 65535)
	    err = 1;
	break;
    default:
	printf("bldRecDef: conv_strng default reached\n");
	printf("file %s line number %d\n%s - RecType=%s\n",
	       plineIn->fname, plineIn->line_no, plineIn->line,
	       ascii_name);
	return (-1);
    }
    if (err == 1) {
	printf("bldRecDef: conv_strng - field out of range\n");
	printf("file %s line number %d\n%s - RecType=%s\n",
	       plineIn->fname, plineIn->line_no, plineIn->line,
	       ascii_name);
	return (-1);
    }
    return (0);
}

/****************************************************************************
	copy dbCommon and current record descriptors to array
****************************************************************************/
/*
 * COPYDESCRIPTORPTRS
 */
static void
copyDescriptorPtrs(pdbCommonList, pdbRecordList, pDestAry)
    SLIST          *pdbCommonList;
    SLIST          *pdbRecordList;
    struct fldDes **pDestAry;
{
    SNODE          *pt;
    SNODE          *list;
    ENODE          *node;
    char           *pSrcAddr;
    int             i = 0;
    /* copy dbCommon descriptor pointers first */
    list = (SNODE *) pdbCommonList;
    if (((SLIST *) list)->count) {
	pt = sllFirst(list);
	while (pt) {
	    node = (ENODE *) pt;
	    pSrcAddr = (void *)&node->desc;
	    memcpy((char *)(pDestAry + i),(char *)&pSrcAddr,sizeof(void *));
	    pt = sllNext(pt);
	    /* increment dest pointer; */
	    i++;
	}
    }
    /* copy record descriptor pointers last */
    list = (SNODE *) pdbRecordList;
    if (((SLIST *) list)->count) {
	pt = sllFirst(list);
	while (pt) {
	    node = (ENODE *) pt;
	    pSrcAddr = (void *)&node->desc;
	    memcpy((char *)(pDestAry + i),(char *)&pSrcAddr,sizeof(void *));
	    pt = sllNext(pt);
	    /* increment dest pointer; */
	    i++;
	}
    }
    return;
}

/****************************************************************************
    Decodes 8th->N field descriptors
****************************************************************************/
/*
 * FIELDDECODE
 */
static int
fieldDecode(plineIn, pInfo)
    struct lineIn  *plineIn;
    struct inLineInfo *pInfo;
{
    int             field_type;
    int             SizeDefined = 0;/* signals lines with defined field size */
    short           dflt_size;
    struct fldDes  *pfield = (struct fldDes *) & pInfo->desc;
    field_type = pfield->field_type;

    /* check for legal field_type */
    if (field_type >= DBF_NTYPES || field_type < 0) {
	printf("bldRecDef: fieldDecode - illegal field_type\n");
	printf("file %s line number %d\n%s - RecType=%s\n",
	       plineIn->fname, plineIn->line_no, plineIn->line,
	       ascii_name);
	return (-1);
    }
    if (field_type == DBF_STRING) {
	/* size followed by promptflag */
	if ((inScan(SCAN_STRING, plineIn, pInfo)) < 0)
	    return (-1);
	if ((inScan(SCAN_PMT_FLAG, plineIn, pInfo)) < 0)
	    return (-1);
	SizeDefined = TRUE;
    } else if (field_type == DBF_CHAR
	       || field_type == DBF_UCHAR
	       || field_type == DBF_SHORT
	       || field_type == DBF_USHORT
	       || field_type == DBF_LONG
	       || field_type == DBF_ULONG
	       || field_type == DBF_FLOAT
	       || field_type == DBF_DOUBLE
	       || field_type == DBF_ENUM) {
	/* initial value followed by promptflag */
	if ((inScan(SCAN_INIT_VAL, plineIn, pInfo)) < 0)
	    return (-1);
	if ((conv_strng(plineIn, pInfo, pInfo->initial, &pfield->initial)) < 0)
	    return (-1);
	if ((inScan(SCAN_PMT_FLAG, plineIn, pInfo)) < 0)
	    return (-1);
	/* range field and cvt_type */
	if (pfield->promptflag > 0)
	    if ((getRangeFld(plineIn, pInfo)) < 0)
		return (-1);
    } else if (field_type == DBF_GBLCHOICE || field_type == DBF_RECCHOICE) {
	/* initial followed by  choice_set index */
	if ((inScan(SCAN_INIT_VAL, plineIn, pInfo)) < 0)
	    return (-1);
	if ((conv_strng(plineIn, pInfo, pInfo->initial, &pfield->initial)) < 0)
	    return (-1);
	if ((inScan(SCAN_CHOICE_SET, plineIn, pInfo)) < 0)
	    return (-1);
	if ((inScan(SCAN_PMT_FLAG, plineIn, pInfo)) < 0)
	    return (-1);
    } else if (field_type == DBF_CVTCHOICE || field_type == DBF_DEVCHOICE) {
	/* initial value followed by promptflag */
	if ((inScan(SCAN_INIT_VAL, plineIn, pInfo)) < 0)
	    return (-1);
	if ((conv_strng(plineIn, pInfo, pInfo->initial, &pfield->initial)) < 0)
	    return (-1);
	if ((inScan(SCAN_PMT_FLAG, plineIn, pInfo)) < 0)
	    return (-1);
    } else if (field_type == DBF_INLINK
	       || field_type == DBF_OUTLINK
	       || field_type == DBF_FWDLINK) {
	if ((inScan(SCAN_PMT_FLAG, plineIn, pInfo)) < 0)
	    return (-1);
    } else if (field_type == DBF_NOACCESS) {
	/* size followed by code to be inserted in .h file */
	if ((inScan(SCAN_NOACCESS, plineIn, pInfo)) < 0)
	    return (-1);
	SizeDefined = TRUE;
    } else {
	printf("bldRecDef: fieldDecode - illegal field_type\n");
	printf("file %s line number %d\n%s - RecType=%s\n",
	       plineIn->fname, plineIn->line_no, plineIn->line,
	       ascii_name);
	return (-1);

    }
    /* if field size not required in line - set to default */
    if (!SizeDefined) {
	set_dflt(field_type, &dflt_size);
	pfield->size = dflt_size;
    }
    return (0);
}

/****************************************************************************
    Processes range and cvt fields
****************************************************************************/
/*
 * GETRANGEFLD
 */
static int
getRangeFld(plineIn, pInfo)
    struct lineIn  *plineIn;
    struct inLineInfo *pInfo;
{
    struct fldDes  *pfield = &pInfo->desc;
    if ((inScan(SCAN_RANGE1, plineIn, pInfo)) < 0)
	return (-1);
    if ((inScan(SCAN_RANGE2, plineIn, pInfo)) < 0)
	return (-1);

    switch (pfield->field_type) {
    case DBF_STRING:
    case DBF_GBLCHOICE:
    case DBF_CVTCHOICE:
    case DBF_RECCHOICE:
    case DBF_DEVCHOICE:
    case DBF_INLINK:
    case DBF_OUTLINK:
    case DBF_FWDLINK:
    case DBF_NOACCESS:
    case DBF_FLOAT:
    case DBF_DOUBLE:
    case DBF_ENUM:
	break;
    case DBF_CHAR:
    case DBF_UCHAR:
    case DBF_SHORT:
    case DBF_USHORT:
    case DBF_LONG:
    case DBF_ULONG:
	/* cvt_type is CT_DECIMAL or CT_HEX */
	if ((inScan(SCAN_CVT_TYPE, plineIn, pInfo)) < 0)
	    return (-1);
	break;
    default:
	printf("bldRecDef: getRangeFld - illegal field_type\n");
	printf("file %s line number %d\n%s - RecType=%s\n",
	       plineIn->fname, plineIn->line_no, plineIn->line,
	       ascii_name);
	return (-1);
	break;
    }
    return (0);
}

/****************************************************************************
    Converts string to value
****************************************************************************/
/*
 * GET_NUM_FROM_STRING
 *
 *	return 0,-1	success,fail
 */
static int
get_num_from_string(pstring, pnum, plineIn)
    char           *pstring;	/* input string */
    double         *pnum;	/* place to put number */
    struct lineIn  *plineIn;
{
    double          number;
    unsigned long   hexnum;
    if ((strncmp(pstring, "0x", 2)) == SAME) {	/* HEX */
	if (sscanf(pstring + 2, "%lx", &hexnum) != 1) {
	    printf("bldRecDef: get_num_from_string -  hex CONVERT FAILED\n");
	    printf("file %s line number %d\n%s - RecType=%s\n",
		   plineIn->fname, plineIn->line_no, plineIn->line,
		   ascii_name);
	    return (-1);
	}
	number = (double) hexnum;
    } else if ((sscanf(pstring, "%lg", &number)) != 1) {
	printf("bldRecDef: get_num_from_string - float CONVERT FAILED\n");
	printf("file %s line number %d\n%s - RecType=%s\n",
	       plineIn->fname, plineIn->line_no, plineIn->line,
	       ascii_name);
	return (-1);

    }
    *pnum = (double) number;
    return (0);
}

/****************************************************************************
    Scan various fields etc.
****************************************************************************/
/*
 * INSCAN
 *
 * return	-1	error
 *		0	OK
 *		1	RECTYP LINE
 *		2	ELEMENT LINE
 */
static int
inScan(scan_cmd, plineIn, pInfo)
    int             scan_cmd;
    struct lineIn  *plineIn;
    struct inLineInfo *pInfo;
{
    int             no_byts;
    int             err = 0;
    int             len;
    char            firstString[BYTES50];

    if (plineIn->pnext == NULL) {
	printf("bldRecDef: Error: No more fields left in line\n");
	printf("file %s line number %d\n%s - RecType=%s\n",
	       plineIn->fname, plineIn->line_no, plineIn->line,
	       ascii_name);
	return (-1);
    }
    switch (scan_cmd) {
    case SCAN_TYPE:
	/* obtain at least 1st 7 chars of line */
	sscanf(plineIn->pnext, "%7s", firstString);
	if ((strcmp("RECTYP", firstString)) == SAME) {

	    if ((sscanf(plineIn->pnext, "%s \"%50[^\"]\"%n",
			pInfo->rectyp_keywd,
			ascii_name,
			&no_byts)) == 2)
		return (RECTYP_DETECTED);
	}
	/* Note: the first 7 common fields are decoded for fldDes line */
	no_byts = 0;
	if ((sscanf(plineIn->pnext, "\"%[^\"]\"%s%hd%hd%hd%hd%hd%n",
		    pInfo->prompt,
		    pInfo->fldname,
		    &pInfo->desc.special,
		    &pInfo->desc.as_level,
		    &pInfo->desc.field_type,
		    &pInfo->desc.process_passive,
		    &pInfo->desc.interest,
		    &no_byts)) == 7) {
	    /* fldname is space filled to FLDNAME_SZ */
	    add_sp(pInfo->fldname, FLDNAME_SZ);
	    memcpy(pInfo->desc.fldname,pInfo->fldname,FLDNAME_SZ);
	    /* prompt is always null terminated */
	    len = strlen(pInfo->prompt);
	    len = (len > PROMPT_SZ - 1) ? PROMPT_SZ - 1 : len;
	    strncpy(pInfo->desc.prompt, pInfo->prompt, PROMPT_SZ - 1);
	    pInfo->desc.prompt[PROMPT_SZ - 1] = '\0';
	    nextInputFld(plineIn, &no_byts);
	    return (ELEMENT_DETECTED);
	}
	err = 1;
	break;
    case SCAN_INIT_VAL:
	/* save as string for now */
	if ((sscanf(plineIn->pnext, "%s%n",
		    pInfo->initial,
		    &no_byts)) != 1)
	    err = 1;
	break;
    case SCAN_STRING:
	if ((sscanf(plineIn->pnext, "%hd%n",
		    &pInfo->desc.size,
		    &no_byts)) != 1)
	    err = 1;
	break;
    case SCAN_NOACCESS:
	if ((sscanf(plineIn->pnext, "%hd %[^\n]%n",
		    &pInfo->desc.size,
		    pInfo->text_code,
		    &no_byts)) != 2)
	    err = 1;
	break;
    case SCAN_CHOICE_SET:
	if ((sscanf(plineIn->pnext, "%hd%n",
		    &pInfo->desc.choice_set,
		    &no_byts)) != 1)
	    err = 1;
	break;
    case SCAN_PMT_FLAG:
	if ((sscanf(plineIn->pnext, "%hd%n",
		    &pInfo->desc.promptflag,
		    &no_byts)) != 1)
	    err = 1;
	break;
    case SCAN_RANGE1:
	if ((sscanf(plineIn->pnext, "%hd%s%n",
		    &pInfo->desc.lowfl,
		    pInfo->varString.lowfl,
		    &no_byts)) == 2) {
	    if (pInfo->desc.lowfl == CON) {
		if ((conv_strng(plineIn, pInfo, pInfo->varString.lowfl,
				&pInfo->desc.range1.value)) < 0)
		    return (-1);
	    }
	    if (pInfo->desc.lowfl == VAR) {
		add_sp(pInfo->varString.lowfl, FLDNAME_SZ);
	    }
	    nextInputFld(plineIn, &no_byts);
	    return (0);
	}
	err = 1;
	break;
    case SCAN_RANGE2:
	if ((sscanf(plineIn->pnext, "%hd%s%n",
		    &pInfo->desc.highfl,
		    pInfo->varString.highfl,
		    &no_byts)) == 2) {
	    if (pInfo->desc.highfl == CON) {
		if ((conv_strng(plineIn, pInfo, pInfo->varString.highfl,
				&pInfo->desc.range2.value)) < 0)
		    return (-1);
	    }
	    if (pInfo->desc.highfl == VAR) {
		add_sp(pInfo->varString.highfl, FLDNAME_SZ);
	    }
	    nextInputFld(plineIn, &no_byts);
	    return (0);
	}
	err = 1;
	break;
    case SCAN_CVT_TYPE:
	if ((sscanf(plineIn->pnext, "%hd%n",
		    &pInfo->desc.cvt_type,
		    &no_byts)) != 1)
	    err = 1;
	break;
    default:
	printf("bldRecDef: inScan - default reached\n");
	err = 1;
	break;
    }
    if (err == 1) {
	printf("bldRecDef: inScan - line format Error\n");
	printf("file %s line number %d\n%s - RecType=%s\n",
	       plineIn->fname, plineIn->line_no, plineIn->line,
	       ascii_name);
	return (-1);
    }
    nextInputFld(plineIn, &no_byts);
    return (0);
}

/****************************************************************************
   Orders the field name and the associated field offset
   in the sortFldName and sortFldInd arrays
****************************************************************************/
/*
 * ORDER_FIELDS
 */
static void
order_fields(precTypDes)
    struct recTypDes *precTypDes;
{
    register short  orderinx;
    short           entry,
                   *pentry;
    char           *pFldName;
    unsigned long  *pFldNameArr;
    short          *pFldIndArr;
    unsigned long  *pLconv;
    short          *pSconv;
    short           no_ordered;
    no_ordered = 0;
    pFldNameArr = precTypDes->sortFldName;
    pFldIndArr = precTypDes->sortFldInd;

    for (entry = 0; entry < precTypDes->no_fields; entry++) {
	pFldName = &precTypDes->papFldDes[entry]->fldname[0];
	pentry = &entry;
	pLconv = pFldNameArr;
	pSconv = pFldIndArr;
	/* find entry place in the order so far */
	orderinx = 0;
	while ((orderinx < no_ordered)
		&& no_ordered > 0
	   && (strncmp((char *)(pLconv - 1), pFldName, FLDNAME_SZ)) > 0) {
	    /* open up the space */
	    memcpy((char *)pLconv,(char *)(pLconv - 1),FLDNAME_SZ);
	    memcpy((char *)pSconv,(char *)(pSconv - 1), sizeof(short));
	    pLconv--;
	    pSconv--;
	    orderinx++;
	}
	/* enter at proper position */
	memcpy((char *)pLconv,pFldName,FLDNAME_SZ);
	memcpy((char *)pSconv,(char *)pentry,sizeof(short));
	pFldNameArr++;
	pFldIndArr++;
	/* increment number ordered */
	no_ordered++;
    }
    return;
}

/****************************************************************************
    Calculate the record summary
****************************************************************************/
/*
 * PLANTRECORDINFO
 */
static void
plantRecordInfo(pdbCommonList, pdbRecordList,
		pno_fields, pno_links, pno_prompt, plast_offset)
    SLIST          *pdbCommonList;
    SLIST          *pdbRecordList;
    int            *pno_fields;
    int            *pno_links;
    int            *pno_prompt;
    int            *plast_offset;
{
    SNODE          *pt;
    SNODE          *list;
    ENODE          *node;
    int             i = 0;
    int             field_type;
    int             last_common = 0;
    int             tmpPmt;
    int             this_offset = 0;
    int             this_size = 0;
    /* calc the dbCommon part first */
    list = (SNODE *) pdbCommonList;

    if (((SLIST *) list)->count) {
	pt = sllFirst(list);
	while (pt) {
	    node = (ENODE *) pt;
	    field_type = node->desc.field_type;
	    if (field_type == DBF_INLINK
		    || field_type == DBF_OUTLINK
		    || field_type == DBF_FWDLINK)
		*pno_links += 1;
	    if (node->desc.promptflag > 0) {
		*pno_prompt += 1;
/*		tmpPmt = node->desc.promptflag; */
/*		last_common = (last_common < tmpPmt ? tmpPmt : last_common); */
	    }
	    pt = sllNext(pt);
	    /* increment dest pointer; */
	    i++;
	}
    }
    /* then calc the Record specific part */
    list = (SNODE *) pdbRecordList;
    if (((SLIST *) list)->count) {
	pt = sllFirst(list);
	while (pt) {
	    node = (ENODE *) pt;
	    field_type = node->desc.field_type;
	    this_offset = node->desc.offset;
	    this_size = node->desc.size;
	    if (field_type == DBF_INLINK
		    || field_type == DBF_OUTLINK
		    || field_type == DBF_FWDLINK)
		*pno_links += 1;
	    if (node->desc.promptflag > 0) {
		*pno_prompt += 1;
/*		node->desc.promptflag += last_common; */
	    }
	    pt = sllNext(pt);
	    i++;
	}
    }
    *pno_fields = i;
    *plast_offset = this_offset + this_size;
    return;
}

/****************************************************************************
    Process range1 and range2 VAR info
****************************************************************************/
/*
 * PLANTVARRANGEINFO
 */
static int
plantVarRangeInfo(pdbCommonList, pdbRecordList, element_count)
    SLIST          *pdbCommonList;
    SLIST          *pdbRecordList;
    int             element_count;

{
    SNODE          *plist;
    SNODE          *pt;
    ENODE          *enode;
    int             cntA;
    int             elementNum = 1;
    int             field_type;
    int             fldnum;
    int             cnt;
    plist = (SNODE *) pdbCommonList;
    cntA = ((SLIST *) pdbCommonList)->count;

    /* for each element - determine ranges */
    for (cnt = 1; cnt < element_count + 1; cnt++) {

	/* is element in dbCommon area */
	if (cnt <= cntA) {
	    elementNum = cnt;	/* YES */
	    plist = (SNODE *) pdbCommonList;
	    pt = sllNth(plist, elementNum);
	}
	if (cnt > cntA) {
	    elementNum = cnt - cntA;
	    plist = (SNODE *) pdbRecordList;
	    pt = sllNth(plist, elementNum);
	}
	enode = (ENODE *) pt;
	field_type = enode->desc.field_type;
	if (field_type == DBF_CHAR
		|| field_type == DBF_UCHAR
		|| field_type == DBF_SHORT
		|| field_type == DBF_USHORT
		|| field_type == DBF_LONG
		|| field_type == DBF_ULONG
		|| field_type == DBF_FLOAT
		|| field_type == DBF_DOUBLE
		|| field_type == DBF_ENUM) {
	    if (enode->desc.lowfl == CON) {
		;		/* already converted */
	    } else if (enode->desc.lowfl == VAR) {
		if ((fldnum = searchForFldnum(pdbCommonList, pdbRecordList,
				element_count, enode->varString.lowfl)) < 0)
		    return (-1);
		enode->desc.range1.fldnum = fldnum;
	    } else {
		printf("bldRecDef: plantVarRangeInfo - illegal value for lowfl\n");
		return (-1);
	    }
	    if (enode->desc.highfl == CON) {
		;		/* already converted */
	    } else if (enode->desc.highfl == VAR) {
		if ((fldnum = searchForFldnum(pdbCommonList, pdbRecordList,
			       element_count, enode->varString.highfl)) < 0)
		    return (-1);
		enode->desc.range2.fldnum = fldnum;
	    } else {
		printf("bldRecDef: plantVarRangeInfo - illegal value for highfl\n");
		return (-1);
	    }
	}
    }
    return (0);
}

/****************************************************************************
    Finds field number of field name
    Returns field number or -1  on error
****************************************************************************/
/*
 * SEARCHFORFLDNUM
 */
static int
searchForFldnum(pdbCommonList, pdbRecordList, element_count, pfldname)
    SLIST          *pdbCommonList;
    SLIST          *pdbRecordList;
    int             element_count;
    char           *pfldname;

{
    SNODE          *plist;
    SNODE          *pt;
    ENODE          *enode;
    int             cntA;
    int             elementNum = 1;
    int             fldNo;
    plist = (SNODE *) pdbCommonList;
    cntA = ((SLIST *) pdbCommonList)->count;

    /* search for field name */
    for (fldNo = 1; fldNo < element_count + 1; fldNo++) {

	/* is element in dbCommon area */
	if (fldNo <= cntA) {
	    elementNum = fldNo;	/* YES */
	    plist = (SNODE *) pdbCommonList;
	    pt = sllNth(plist, elementNum);
	}
	if (fldNo > cntA) {
	    elementNum = fldNo - cntA;
	    plist = (SNODE *) pdbRecordList;
	    pt = sllNth(plist, elementNum);
	}
	enode = (ENODE *) pt;
	if ((strncmp(enode->desc.fldname, pfldname, 4)) == SAME)
	    return (fldNo - 1);
    }
    printf("bldRecDef: searchForFldnum - could not find fldname %4s\n", pfldname);
    return (-1);
}

/****************************************************************************
   Sets the default size for each type
****************************************************************************/
/*
 * SET_DFLT
 */
static void
set_dflt(field_type, pdflt_size)
    short           field_type;
    short          *pdflt_size;
{
    /* define default sizes */
    switch (field_type) {
    case DBF_CHAR:
    case DBF_UCHAR:
	*pdflt_size = sizeof(char);
	break;
    case DBF_SHORT:
    case DBF_USHORT:
	*pdflt_size = sizeof(short);
	break;
    case DBF_LONG:
    case DBF_ULONG:
	*pdflt_size = sizeof(long);
	break;
    case DBF_FLOAT:
	*pdflt_size = sizeof(float);
	break;
    case DBF_DOUBLE:
	*pdflt_size = sizeof(double);
	break;
    case DBF_ENUM:
    case DBF_GBLCHOICE:
    case DBF_CVTCHOICE:
    case DBF_RECCHOICE:
    case DBF_DEVCHOICE:
	*pdflt_size = sizeof(short);
	break;
    case DBF_INLINK:
    case DBF_OUTLINK:
    case DBF_FWDLINK:
	*pdflt_size = sizeof(struct link);
	break;
    default:
	break;
    }
    return;
}

/****************************************************************************
    Verify record type
****************************************************************************/
/*
 * VERIFYTYPE
 */
static int
verifyType(dbRecType, pRec_no)
    struct recType *dbRecType;
    int            *pRec_no;
{
    short           i;
    /* Verify record type name */
    for (i = 0; i < dbRecType->number; i++) {
	if (dbRecType->papName[i] != NULL) {
	    if (strcmp(dbRecType->papName[i],
		       ascii_name) == SAME) {
		*pRec_no = i;
		return (TRUE);
	    }
	}
    }
    printf("bldRecDef: verifyType - Record type '%s' not defined\n",
	   ascii_name);
    return (FALSE);
}
/****************************************************************************
    Writes buffer to file
****************************************************************************/
/*
 * WRT_BUF
 */
static int
wrt_buf(fp, buf, pfname)
    FILE           *fp;
    char           *pfname;
    char           *buf;
{
    int             rtnval;
    int             len;
    len = strlen(buf);
    if ((rtnval = fwrite(buf, len, 1, fp)) != 1) {
	printf("bldRecDef: fwrite error. errno=%i rtnval=%i\nfile=%s.h\n",
	       errno, rtnval, pfname);
	return (-1);
    }
    return (0);
}

/* top common part of all header files */
static char    *text_str[] = {
    "#include <ellLib.h>	/* list structure definition for the monitor list */",
    "#ifndef	INCLsemLibh",
    "#include <semLib.h>	/* semaphore structure definition for the monitor lock*/",
    "#endif",
    "#ifndef INCLfast_lockh",
    "#include <fast_lock.h>	/* fast lock data structures and macros */",
    "#endif",
    "#ifndef INCLlinkh",
    "#include <link.h>	/* fast lock data structures and macros */",
    "#endif",
    "#ifndef INC_tsDefs_h",
    "#include <tsDefs.h>	/* time-stamp related definitions */",
    "#endif"
};
#define TEXT_COUNT (sizeof(text_str) / sizeof(void *))
/****************************************************************************
    Writes common part of a header file
****************************************************************************/
/*
 * WRT_COM
 */
static int
wrt_com(fp, buf, pfname, psname)
    FILE           *fp;
    char           *pfname;
    char           *psname;
    char           *buf;
{
    char          **ptr = text_str;
    short           i;
    /* 1st 3 lines of header */
    sprintf(buf, "\n/* %s */", pfname);
    if ((wrt_buf(fp, buf, pfname)) < 0)
	return (-1);
    sprintf(buf, "\n#ifndef INC%sh", psname);
    if ((wrt_buf(fp, buf, pfname)) < 0)
	return (-1);
    sprintf(buf, "\n#define INC%sh", psname);
    if ((wrt_buf(fp, buf, pfname)) < 0)
	return (-1);

    /* common text of header */
    for (i = 0; i < TEXT_COUNT; i++, ptr++) {
	strcpy(buf, "\n");
	strcat(buf, *ptr);
	if ((wrt_buf(fp, buf, pfname)) < 0)
	    return (-1);
    }
    /* variable start of struct name */
    strcpy(buf, "\nstruct ");
    strcat(buf, psname);
    strcat(buf, "	{");
    if ((wrt_buf(fp, buf, pfname)) < 0)
	return (-1);
    return (0);
}
