/*  $Id: getSdrLine.c,v 1.3 1991/06/07 18:24:35 rcz Exp $
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02	mm-dd-yy	iii	Comment
 * 	...
 */

#include <stdio.h>
#include <ctype.h>
#include <getSdrLine.h>

#ifndef TRUE
#define TRUE            1
#endif
#ifndef FALSE
#define FALSE           0
#endif

/*
 * GETLINE
 *
 * Reads the next line from an SDR ascii file. Removes comments and
 * if line is empty - repeats.  Counts the total number of double
 * quotes in the output line and reports various errors.
 *
 * struct lineIn
 *         *pnext           pointer to next field in line[]
 *                                 NULL if no more fields
 *         line[]           input line buffer
 *         quote_count      number of double quotes in line
 *         line_no          current line number
 *         eof              set to EOF or 0
 *         *fname           pointer to input filename
 *
 * The structure must be allocated and elements fname and line_no
 * initialized prior to the 1st call. Upon exit, line_no is incremented,
 * eof set to EOF or 0, quote_count is set and pnext set to line[0].
 *
 * return	-1	error
 * return	0	success
 *
 */
getline(fp, pline)
    FILE           *fp;
    struct lineIn  *pline;
{
    int             status;
    /* get next line from file */
    FOREVER {
	pline->line_no++;
	status = gline(fp, pline);
	if ((gStat(status, pline)) < 0)
	    return (-1);
	if (pline->eof == EOF)
	    break;
	if ((pline->line[0]) == '\n')
	    continue;
	break;
    }
    return (0);
}
/*
 * GLINE
 * part of getline
 *
 */
gline(fp, pline)
    FILE           *fp;
    struct lineIn  *pline;
{
    register        c,
                    i;
    static int      lastchar;
    int             erase;
    int             non_print;
    int             lim;
    i = erase = non_print = 0;
    lim = MAXLINE;

    pline->quote_count = 0;
    pline->eof = 0;

    pline->pnext = pline->line;

    while (--lim > 0 && (c = getc(fp)) != EOF && c != '\n') {
	if (i == 0 && (c == ' ' || c == '\t'))
	    continue;
	if (c == '!')
	    erase = 1;
	/* process the rest of the line */
	if (!erase) {
	    if (c == '"')
		pline->quote_count++;
	    if (!(isprint(c) || isspace(c)))
		non_print = 1;
	    pline->pnext[i++] = c;
	}
	lastchar = c;
    }
    pline->pnext[i++] = '\n';
    pline->pnext[i] = '\0';
    if (lim == 0)		/* line too long */
	return (3);
    if (non_print)
	return (1);
    if (c == '\n') {
	lastchar = '\n';
	return (0);
    }
    if (c == EOF && lastchar == '\n') {
	pline->eof = EOF;
	return (0);		/* normal EOF */
    }
    return (2);			/* abnormal EOF */
}
/*
 * GSTAT
 *
 * part of getline
 * prints any status errors
 *
 * return -1  error
 * return  0  success
 */
gStat(status, pline)
    struct lineIn  *pline;
    int             status;
{

    if (pline->quote_count & 0x1) {
	printf("gStat(): Odd number of double quotes on line %i \n", pline->line_no);
	return (-1);
    }
    switch (status) {
    case 0:
	return (0);
    case 1:
	printf("gStat(): Non-printable character on line %i \n", pline->line_no);
	break;
    case 2:
	printf("gStat(): Abnormal EOF on line %i \n", pline->line_no);
	break;
    case 3:
	printf("gStat(): Max line length exceeded on line %i \n", pline->line_no);
	break;
    default:
	printf("gStat(): Unknown status code =%i\n", status);
	break;
    }
    return (-1);
}
/*
 * GET_SDR_LINE
 *
 * Read the next SDR line and expect nquotes double quotes.
 *
 * return	-1	error
 * return	0	success
 *
 * Note:  EOF must be tested by caller.
 */
getSdrLine(fp, plineIn, nquotes)
    FILE           *fp;
    struct lineIn  *plineIn;
    int             nquotes;
{
    if ((getline(fp, plineIn)) < 0)
	return (-1);
    if (plineIn->eof == EOF)
	return (0);
    /* lines must have exactly  nquotes double quotes */
    if (plineIn->quote_count != nquotes) {
	printf("getSdrLine: Error - line=%i: Wrong number of double quotes\n",
	       plineIn->line_no);
	return (-1);
    }
    return (0);
}


/*
 * NEXTINPUTFLD
 */
nextInputFld(plineIn, pno_byts)
    struct lineIn  *plineIn;
    int            *pno_byts;
{
    char           *Wptr;
    int             anotherfield;
    /* step to next field */
    plineIn->pnext += *pno_byts;
    /* test rest of line empty */
    Wptr = plineIn->pnext;

    if (*Wptr == '\n')
	plineIn->pnext = NULL;

    anotherfield = FALSE;
    while (*Wptr != '\n') {
	if (isspace(*Wptr++)) {
	    plineIn->pnext += 1;
	    continue;
	} else {
	    anotherfield = TRUE;
	    break;
	}
    }
    if (!anotherfield)
	plineIn->pnext = NULL;
}
