/*  $Id: dbReadWrite.c,v 1.9 1995/02/28 22:04:40 mrk Exp $
 *
 *	Original Authors: Bob Dalesio, Julie Sander (Original DCT)
 *	Second Author:   Frank Lenkszus (Changes to support extensible record support)
 *	Current Author:  Robert Zieman (Access Layer for .database files)
 *      Date:	07-20-92
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	04-08-92	rcz	initial development
 * .02  07-20-92        rcz     revised header doc
 * .03  08-31-92        rcz     suppressed checksum from dbRead for DCT
 * .04  09-16-92        rcz     added fatal_errors to OPENFILE case
 * .05  12-24-92        rcz     fixed bug in initDbRead and some cleanup
 * .06  02-03-93        rcz     re-fixed bug in initDbRead
 *
 */

/* Loads self defining records */

/* An input file to be loaded by dbRead consists
 * of one or more SDR files. SDR_DB_RECTYPE is needed to decipher many
 * of the other SDRs. If it is present, it must be the first SDR in
 * the input file.
 */

/* Note: dbWrite can't be used in vxWorks at present */

/******************************************************************
 * DBREAD - loads one or more databases into memory
 *
 *    adds records to the pvd list and the appropriate record list
 *
 *    creates a sorted record list for each record type
 *
 *    An SDR record is composed of a header "type" followed by a data block
 *
 *    Format of a database:
 *        =========================================================
 *        
 *        =========================================================
 *        The set of SDR record "types" contained in the
 *             default.dctsdr file.
 *        "type" SDR_DB_RECTYPE - must be the first SDR record in the set.
 *                ...           - other ones in the set.
 *        "type" SDR_ALLSUMS    - is the last SDR record in the set.
 *        =========================================================
 *        Followed by one or more SDR records (of type SDR_DB_RECORDS)
 *            SDR_DB_RECORDS
 *                ...            - one for each recType in database
 *        =========================================================
 *        =========================================================
 *
 *      Each SDR record of type SDR_DB_RECORDS contains a
 *      recLoc structure followed by the contiguos set of PV records
 *      for the record type defined in the recLoc structure.
 *
 *      The 1st database loaded by dbRead creates and initializes various
 *      structures and computes the sdrFileSize of the default.dctsdr file
 *      which was copied verbatim when the database was created.
 *      This is currently only used when writing the database on the host.
 *      All databases that are loaded must have been created from the
 *      same default.dctsdr file and is enforced by dbRead.
 *
 *      When an attempt is made to load a PV already in the database
 *      a warning message is issued, and if the record type agrees
 *      the current version will be overwritten. Otherwise the new
 *      PV is skipped.
 *
 *      The definition of the 1st database:
 *          start of loading - arg1  (ppdbBase) is NULL
 *          end - EOF on file (and SDR_ALLSUMS found and checked)
 *          If ppdbBase != NULL it is NOT the 1st database.
 *       The in-memory database name is the name of the 1st one loaded, however
 *       the name of each is output as it is loaded.
 *****************************************************************/



#ifdef vxWorks
#include <vxWorks.h>
#include <string.h>
#include <types.h>
#include <ctype.h>
#include <errnoLib.h>
#include <stdioLib.h>
#include <stat.h>
#include <memLib.h>
extern VOID    *calloc();
#define P_TYPE	unsigned
extern int      errMessage();
#else

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <dbDefs.h>
#define P_TYPE	int
#endif


#include <ellLib.h>
#include <dbBase.h>
#include "dbManipulate.h"
#include <sdrHeader.h>
#include <choice.h>
#include <dbRecType.h>
#include <dbRecDes.h>
#include <dbRecords.h>
#include <devSup.h>
#include <drvSup.h>
#include <cvtTable.h>
#include <errMdef.h>

#ifndef FOREVER
#define FOREVER     for(;;)
#endif
#if	!defined(FALSE) || (FALSE!=0)
#define FALSE		0
#endif
#if	!defined(TRUE) || (TRUE!=1)
#define TRUE		1
#endif
#define SAME 0

/*Convert an offset value to an address - taken from dbStaticLib.c */
#ifdef __STDC__
static void *cvtOffsetToAddr(void *pbase,void *poff)
#else
static void *cvtOffsetToAddr(pbase,poff)
void *pbase;
void *poff;
#endif /*__STDC__*/
{
    long   offset;
    void   **poffset= (void **)&offset;

    *poffset = poff;
    return((void *)((char *)pbase + offset));
}

/* forward references */
static int	adj_dbRecType();
static int	adj_dbRecords();
static int	adj_dbRecDes();
static int	adj_dbPvd();
static int	adj_choiceGbl();
static int	adj_choiceCvt();
static int	adj_choiceRec();
static int	adj_choiceDev();
static int	adj_devSup();
static int	adj_cvtTable();
static int	adj_drvSup();
static int	adj_recSup();
static int	adj_sdrSum();
static int	initDbRead();
static int	getFileName();
static int	addValidDefaults();
static int	checkSdrSum();

/* array of pointers to adjusting functions */
static          (*adj_fun[SDR_NTYPES]) ();

/* Note: see addValidDefaults() function for sdr types in default.dctsdr file */

static long     i, j, k;
static long     status;
static char     text[120];


/* states */
#define OPENFILE	0
#define READHEADER	1
#define READDATA	2
#define SKIPDATA	3
#define GETFILENAME	4
#define CLOSEFILE	5
#define ENDJOB		6
#define ENDFILENAMES	7
#define INSTALLDATA	8

/* job types */
#define  FIND_SUM	1
#define  LOAD_ONLY_ONE	2	/* only for bld tools */
#define  LOAD_DATABASE	3
#define  LOAD_DCT_DEFAULTS	4

#define MAX_DBREAD_FILENAME 256

static char     current_file[MAX_DBREAD_FILENAME];
static FILE    *sdr_fp = NULL;
static int      state = GETFILENAME;
static int      fileCtrlNumber;	/* controls type of processing */
static char    *iptr = NULL;

static struct sdrHeader sdrHeader; /* place to read in SDR headers */


/*
 * MAIN FOR DBREAD
 *
 * Rework of sdrLoad to allow access to databases independent of DCT.
 * Allows multiple databases to be put in memory with one call.
 *
 * input   address of a pointer to pointer to dbBase
 *         Note: if the pointer is NULL, dbBase is allocated.
 *
 * return    status  0 = success, else failure
 *
 */

#ifdef __STDC__
long dbRead(struct dbBase **ppdbBase, char *pdatabaseName)
#else
long
dbRead(ppdbBase, pdatabaseName)
    struct dbBase **ppdbBase;
    char           *pdatabaseName;	/* space seperated filename(s) */
#endif
{
    struct dbBase  *pdbBase;
    int             fatal_errors = 0;
    int             len;
    char            files_line[MAX_DBREAD_FILENAME];
    long            offset;
    int             job_type = 999;
    long            sdrFileSize = 0;	/* size of default.dctsdr file */
    char           *ptmp = NULL;
    int             sums_checked = 0;
    struct sdrSum  *psdrSum = NULL;
    len = strlen(pdatabaseName);
    iptr = NULL;
    if (len >= MAX_DBREAD_FILENAME - 1) {
	sprintf(text, "dbRead: arg pdatabaseName string too long %s", current_file);
	errMessage(-1L, text);
	return (-1);
    }
    memcpy(files_line, pdatabaseName, len + 1);
    if ((initDbRead(ppdbBase, pdatabaseName)) < 0)
	return (-1);
    pdbBase = *ppdbBase;
    /* determine type of file to process */
    if (pdbBase->pvtSumFlag == 0) {
	fileCtrlNumber = 1;	/* starts at 1 for first database */
	job_type = FIND_SUM;	/* prior to loading the database */
    } else if (pdbBase->pvtSumFlag == 1 )		 {
	fileCtrlNumber = 0;
	/* load only one SDR file w/o checking sum */
	job_type = LOAD_ONLY_ONE;
    } else if (pdbBase->pvtSumFlag == 2) {
	fileCtrlNumber = 0;
	/* load all of the SDR files in default.dctsdr w/o checking sum */
	job_type = LOAD_DCT_DEFAULTS;
    }
    state = GETFILENAME;


    while (1) {
	switch (state) {
	    case
	GETFILENAME:
	    if ((getFileName(files_line)) < 0) {
		/* all databases processed */
		state = ENDFILENAMES;
		break;
	    }
	    psdrSum = NULL;
	    sums_checked = 0;
	    state = OPENFILE;
	    break;
	    case
	OPENFILE:
	    if ((sdr_fp = fopen(current_file, "r")) == NULL) {
		status = S_sdr_noOpen;
		sprintf(text, "dbRead: Can't fopen file=(%s)", current_file);
		errMessage(status, text);
		fatal_errors = 1;
		state = GETFILENAME;
		break;
	    }
	    state = READHEADER;
	    break;


	    case
	READHEADER:		/* reads all sdr header types */
	    if ((fread((char *) &sdrHeader, sizeof(struct sdrHeader),
		       1, sdr_fp)) != 1) {
		if (feof(sdr_fp)) {
		    /* normal termination of a file */
		    fclose(sdr_fp);
		    if (!sums_checked && job_type == LOAD_DATABASE) {
			sprintf(text,
				"dbRead: READHEADER - LOAD_DATABASE - checkSdrSum() not invoked ");
			errMessage(-1L, text);
			fatal_errors = 1;
			state = CLOSEFILE;
			break;
		    }
		    if (job_type == LOAD_DCT_DEFAULTS) {
			state = ENDFILENAMES;
			break;
		    }
		    state = GETFILENAME;
		    if (job_type == LOAD_DATABASE)
			job_type = FIND_SUM;
		    break;
		}
		status = S_sdr_noRead;
		sprintf(text, "dbRead: READHEADER - fread of %s", current_file);
		errMessage(status, text);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    }
	    if (sdrHeader.magic != DBMAGIC) {
		status = S_sdr_noMagic;
		sprintf(text, "dbRead: READHEADER - wrong DBMAGIC %s", current_file);
		errMessage(status, text);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    }
	    if (sdrHeader.type > SDR_NTYPES || sdrHeader.type < 0) {
		status = S_sdr_noRecDef;
		sprintf(text, "dbRead: READHEADER - type=%d > SDR_NTYPES  ", sdrHeader.type);
		errMessage(status, text);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    }
	    if (job_type == FIND_SUM && sdrHeader.type == SDR_DB_RECORDS) {
		sprintf(text,
		"dbRead: READHEADER - SDR_DB_RECORDS found during FIND_SUM");
		errMessage(-1L, text);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    }
	    if (job_type == FIND_SUM) {
		/* calculate default.dctsdr file size  from first file only*/
		if (fileCtrlNumber == 1) {
		    addValidDefaults(&sdrFileSize, &sdrHeader);
		    if (sdrHeader.type == SDR_ALLSUMS) {
			if (!pdbBase->sdrFileSize) {
			    pdbBase->sdrFileSize = sdrFileSize;
			}
			state = READDATA;
			break;
		    }
		}
		state = SKIPDATA;
		break;
	    }
	    if (job_type == LOAD_ONLY_ONE) {
		if (sdrHeader.type == SDR_DB_RECORDS) {
		    sprintf(text,
			    "dbRead: READHEADER - SDR_DB_RECORDS header found during LOAD_ONLY_ONE");
		    errMessage(-1L, text);
		    fatal_errors = 1;
		    state = CLOSEFILE;
		    break;
		}
		state = READDATA;
		break;
	    } else if (!sums_checked && job_type == LOAD_DATABASE) {
		sprintf(text,
			"dbRead: READHEADER - LOAD_DATABASE - checkSdrSum() not invoked ");
		errMessage(-1L, text);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    } else if (sums_checked && job_type == LOAD_DATABASE) {
		state = READDATA;
		break;
	    } else if (job_type == LOAD_DCT_DEFAULTS) {
		state = READDATA;
		break;
	    } else {
		sprintf(text,
			"dbRead: READHEADER - unknown job_type=%d %s ", job_type, current_file);
		errMessage(status, text);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    }
	    break;


	    case
	SKIPDATA:
	    /* seek offset bytes from current postion */
	    offset = sdrHeader.nbytes;
	    if ((fseek(sdr_fp, offset, 1L)) != 0) {
		status = S_sdr_noRead;
		sprintf(text,
			"dbRead: SKIPDATA - incomplete seek %s job_type=%d", current_file, job_type);
		errMessage(status, text);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    }
	    state = READHEADER;
	    break;
	    case
	CLOSEFILE:
	    sprintf(text, "dbRead: Aborting file %s", current_file);
	    errMessage(-1L, text);
	    fclose(sdr_fp);
	    state = ENDJOB;
	    break;


	    case
	READDATA:
	    /* allocate space for the SDR data block */
	    if ((ptmp = (char *) calloc((unsigned) 1, (unsigned) sdrHeader.nbytes)) == NULL) {
		status = S_sdr_noAlloc;
		sprintf(text, "dbRead: calloc error ");
		errMessage(status, text);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
		/* load sdr data block */
	    }
	    if ((fread((char *) ptmp, (P_TYPE) sdrHeader.nbytes, 1, sdr_fp)) != 1) {
		free(ptmp);
		status = S_sdr_noRead;
		sprintf(text, "dbRead: Error reading file %s", current_file);
		errMessage(status, text);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    }
	    if (job_type == FIND_SUM && sdrHeader.type == SDR_ALLSUMS) {
		psdrSum = (struct sdrSum *) ptmp;
		if ((checkSdrSum(psdrSum)) < 0) {
		    free(ptmp);
		    fatal_errors = 1;
		    state = CLOSEFILE;
		    break;
		} else {
		    free(ptmp);
		    sums_checked = 1;
		    job_type = LOAD_DATABASE;
		    rewind(sdr_fp);
		    state = READHEADER;
		    break;
		}
	    }
	    if (job_type == FIND_SUM) {
		sprintf(text,
			"dbRead: READDATA - job_type FIND_SUM failed");
		errMessage(-1L, text);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    }
/*******##################################### */
	    if (job_type == LOAD_DATABASE 
		|| job_type == LOAD_ONLY_ONE
		|| job_type == LOAD_DCT_DEFAULTS) {
		state = INSTALLDATA;
	    } else {
		sprintf(text, "dbRead: READDATA - internal error %s ", current_file);
		errMessage(status, text);
		fatal_errors = 1;
		state = CLOSEFILE;
	    }
	    break;

	    case
	INSTALLDATA:
	    if ((adj_fun[sdrHeader.type]) == NULL) {
		status = S_sdr_noSdrType;
		sprintf(text, "dbRead: Warning: S_sdr_noSdrType - type=%d not found",
			sdrHeader.type);
		errMessage(status, text);
		free(ptmp);
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    }
	    /* pass the adjusting function the allocated ptr */
	    if (((*adj_fun[sdrHeader.type]) (ptmp, pdbBase)) < 0) {
		fatal_errors = 1;
		state = CLOSEFILE;
		break;
	    }
	    if (job_type == LOAD_ONLY_ONE) {
		state = ENDFILENAMES;
		break;
	    }
	    state = READHEADER;
	    break;
	    case
	ENDJOB:
	    /* should never get here */
	    break;
	default:
	    sprintf(text, "dbRead: internal error - Unknown state=%d ", state);
	    errMessage(-1L, text);
	    fatal_errors = 1;
	    state = CLOSEFILE;
	    break;
	}			/* end switch */
	if (state == ENDJOB) {
	    state = GETFILENAME;
	    if (job_type = LOAD_DATABASE)
		job_type = FIND_SUM;
	}
	if (state == ENDFILENAMES)
	    break;
    }
    if (fatal_errors)
	return (-1);
    return (0);
}

/*
 * ADJ_DBRECTYPE
 * 
 * Load and adjust SDR structure
 * 
 */
static int 
adj_dbRecType(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    struct recType *precType = NULL;

    if (pdbBase->precType == NULL) {
	precType = (struct recType *) ptmp;
	precType->papName = cvtOffsetToAddr(precType,precType->papName);
	for (i = 0; i < precType->number; i++) {
	    if (precType->papName[i] != NULL)
		precType->papName[i] = cvtOffsetToAddr(precType,
			precType->papName[i]);
	}
	pdbBase->precType = precType;
    } else {
	/* ignore - already loaded */
	free(ptmp);
    }
    return 0;
}

/*
 * ADJ_DBRECORDS
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_dbRecords(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    DBENTRY         dbEntry;
    struct recHeader *precHeader = (struct recHeader *) pdbBase->precHeader;
    struct recType *precType = (struct recType *) pdbBase->precType;
    struct recDes  *precDes = (struct recDes *) pdbBase->precDes;
    /* set ptr to first of contiguous set of records in ptmp */
    char           *precord = (char *) ptmp + sizeof(struct recLoc);
    struct recLoc  *precLoc = (struct recLoc *) ptmp;
    long            no_records;
    long            offset;
    int             found = 0;
    if (precType == NULL) {
	status = S_sdr_notLoaded;
	sprintf(text, "dbRead: adj_dbRecords() recType == NULL");
	errMessage(status, text);
	return (-1);
    }
    if (precDes == NULL) {
	status = S_sdr_notLoaded;
	sprintf(text, "dbRead: adj_dbRecords() recDes == NULL");
	errMessage(status, text);
	return (-1);
    }
    /* if 1st database - init recHeader */
    if (precHeader == NULL) {
	offset = sizeof(struct recHeader) + precType->number * sizeof(char *);
	if ((precHeader = (struct recHeader *)
	     calloc((unsigned) 1, (unsigned) offset)) == NULL) {
	    status = S_sdr_noAlloc;
	    sprintf(text, "dbRead: adj_dbRecords() calloc error == NULL");
	    errMessage(status, text);
	    return (-1);
	}
	precHeader->papRecLoc = (struct recLoc **)(precHeader) + 1;
	precHeader->number = precType->number;
	pdbBase->precHeader = precHeader;
    }
    no_records = (sdrHeader.nbytes / precLoc->rec_size);
    for (i = 0; i < no_records; i++, precord += precLoc->rec_size) {
	if (*precord == '\0')
	    continue;
	found = 0;
	if ((dbFindRecord(pdbBase, (char *) precord, &dbEntry)) == 0) {
	    found = 1;
	}
	/* record found - does type match */
	if (found && (dbEntry.record_type != precLoc->record_type)) {
		sprintf(text,
		 "dbRead: adj_dbRecords: record (%s) skipped - wrong type ",
			 (char *) precord);
		errMessage(-1L, text);
	} else if (found && (dbEntry.record_type == precLoc->record_type)) {
	    /* copy precord to dbEntry.precord */
	    memcpy( (char *) dbEntry.precord, (char *) precord , precLoc->rec_size);
		sprintf(text,
		 "dbRead: adj_dbRecords: record (%s) overwritten ",
			 (char *) precord);
		errMessage(-1L, text);
	} else if (!found) {
	    if ((dbCreateRecord(pdbBase, precLoc->record_type, (char *) precord, &dbEntry)) != 0) {
		sprintf(text, "dbRead: adj_dbRecords: dbCreateRecord  failed");
		errMessage(-1L, text);
	    } else {
	         memcpy((char*)dbEntry.precord, (char*)precord, precLoc->rec_size);
	    }
	}
    }
    free(ptmp);
    return 0;
}

/*
 * ADJ_DBRECDES
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_dbRecDes(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    struct recDes *precDes = (struct recDes*)pdbBase->precDes;
    if (pdbBase->precType == NULL) {
	status = S_sdr_notLoaded;
	sprintf(text, "dbRead: adj_dbRecDes() precType == NULL");
	errMessage(status, text);
	return (-1);
    }
    if (pdbBase->precDes == NULL) {
	precDes = (struct recDes *) ptmp;
	precDes->papRecTypDes = cvtOffsetToAddr(precDes,precDes->papRecTypDes);
	for (i = 0; i < precDes->number; i++) {
	    if (precDes->papRecTypDes[i] != NULL) {
		precDes->papRecTypDes[i] =
		   cvtOffsetToAddr(precDes,precDes->papRecTypDes[i]);
		precDes->papRecTypDes[i]->papFldDes =
		   cvtOffsetToAddr(precDes,precDes->papRecTypDes[i]->papFldDes);
		precDes->papRecTypDes[i]->sortFldName =
		   cvtOffsetToAddr(precDes,precDes->papRecTypDes[i]->sortFldName);
		precDes->papRecTypDes[i]->sortFldInd =
		   cvtOffsetToAddr(precDes,precDes->papRecTypDes[i]->sortFldInd);
		precDes->papRecTypDes[i]->link_ind =
		   cvtOffsetToAddr(precDes,precDes->papRecTypDes[i]->link_ind);
		for (j = 0; j < precDes->papRecTypDes[i]->no_fields; j++) {
		    if (precDes->papRecTypDes[i]->papFldDes != NULL) {
			if (precDes->papRecTypDes[i]->papFldDes[j] != NULL) {
			    precDes->papRecTypDes[i]->papFldDes[j] =
				cvtOffsetToAddr(precDes,precDes->papRecTypDes[i]->papFldDes[j]);
#if 0
			    (long) precDes->papRecTypDes[i]->papFldDes[j]
				+= (long) precDes;
#endif
			}
		    }
		}
	    }
	}
	pdbBase->precDes = precDes;
    } else {
	/* ignore - already loaded */
	free(ptmp);
    }
    return 0;
}

/*
 * ADJ_CHOICEGBL
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_choiceGbl(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    struct arrChoiceSet *pchoiceGbl = NULL;
    if (pdbBase->pchoiceGbl == NULL) {
	pchoiceGbl = (struct arrChoiceSet *) ptmp;
	pchoiceGbl->papChoiceSet = cvtOffsetToAddr(pchoiceGbl,pchoiceGbl->papChoiceSet);
	for (i = 0; i < pchoiceGbl->number; i++) {
	    pchoiceGbl->papChoiceSet[i] = cvtOffsetToAddr(pchoiceGbl,
		pchoiceGbl->papChoiceSet[i]);
	    if (pchoiceGbl->papChoiceSet[i] != NULL) {
		pchoiceGbl->papChoiceSet[i]->papChoice = cvtOffsetToAddr(
		    pchoiceGbl,pchoiceGbl->papChoiceSet[i]->papChoice);
		for (j = 0; j < pchoiceGbl->papChoiceSet[i]->number; j++) {
		    if (pchoiceGbl->papChoiceSet[i]->papChoice[j] != NULL) {
			pchoiceGbl->papChoiceSet[i]->papChoice[j] = cvtOffsetToAddr(
			    pchoiceGbl,pchoiceGbl->papChoiceSet[i]->papChoice[j]);
		    }
		}
	    }
	}
	pdbBase->pchoiceGbl = pchoiceGbl;
    } else {
	/* ignore - already loaded */
	free(ptmp);
    }
    return 0;
}

/*
 * ADJ_CHOICECVT
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_choiceCvt(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    struct choiceSet *pchoiceCvt = NULL;
    if (pdbBase->pchoiceCvt == NULL) {
	pchoiceCvt = (struct choiceSet *) ptmp;
	pchoiceCvt->papChoice = cvtOffsetToAddr(pchoiceCvt,pchoiceCvt->papChoice);
	for (i = 0; i < pchoiceCvt->number; i++) {
	    if (pchoiceCvt->papChoice[i] != NULL) {
		pchoiceCvt->papChoice[i] = cvtOffsetToAddr(pchoiceCvt,
			pchoiceCvt->papChoice[i]);
	    }
	}
	pdbBase->pchoiceCvt = pchoiceCvt;
    } else {
	/* ignore - already loaded */
	free(ptmp);
    }
    return 0;
}

/*
 * ADJ_CHOICEREC
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_choiceRec(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    struct choiceRec *pchoiceRec = NULL;
    struct recType *precType = pdbBase->precType;
    if (pdbBase->precType == NULL) {
	status = S_sdr_notLoaded;
	sprintf(text, "dbRead: adj_choiceRec() recType == NULL");
	errMessage(status, text);
	return (-1);
    }
    if (pdbBase->pchoiceRec == NULL) {
	pchoiceRec = (struct choiceRec *) ptmp;
	pchoiceRec->papArrChoiceSet = cvtOffsetToAddr(pchoiceRec,
		pchoiceRec->papArrChoiceSet);
	for (i = 0; i < precType->number; i++) {
	    if (pchoiceRec->papArrChoiceSet[i] != NULL) {
		pchoiceRec->papArrChoiceSet[i] = cvtOffsetToAddr(
		    pchoiceRec,pchoiceRec->papArrChoiceSet[i]);
		pchoiceRec->papArrChoiceSet[i]->papChoiceSet = cvtOffsetToAddr(
		    pchoiceRec,pchoiceRec->papArrChoiceSet[i]->papChoiceSet);
		for (j = 0; j < pchoiceRec->papArrChoiceSet[i]->number; j++) {
		    if (pchoiceRec->papArrChoiceSet[i]->papChoiceSet[j] != NULL) {
			pchoiceRec->papArrChoiceSet[i]->papChoiceSet[j] = cvtOffsetToAddr(
			    pchoiceRec,pchoiceRec->papArrChoiceSet[i]->papChoiceSet[j]);
			pchoiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice =
			    cvtOffsetToAddr(pchoiceRec,
				pchoiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice);
			for (k = 0; k < pchoiceRec->papArrChoiceSet[i]->papChoiceSet[j]->number; k++) {
			    if (pchoiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k] != NULL) {
				pchoiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k] =
				    cvtOffsetToAddr(pchoiceRec,
				    pchoiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k]);
			    }
			}
		    }
		}
	    }
	}
	pdbBase->pchoiceRec = pchoiceRec;
    } else {
	/* ignore - already loaded */
	free(ptmp);
    }
    return 0;
}

/*
 * ADJ_CHOICEDEV
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_choiceDev(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    struct devChoiceRec *pchoiceDev = (struct devChoiceRec*)pdbBase->pchoiceDev;
    struct recType *precType = (struct recType*)pdbBase->precType;

    if (pdbBase->precType == NULL) {
	status = S_sdr_notLoaded;
	sprintf(text, "dbRead: adj_choiceDev() precType == NULL");
	errMessage(status, text);
	return (-1);
    }
    if (pdbBase->pchoiceDev == NULL) {
	pchoiceDev = (struct devChoiceRec *) ptmp;
	pchoiceDev->papDevChoiceSet = cvtOffsetToAddr(pchoiceDev,pchoiceDev->papDevChoiceSet);
	for (i = 0; i < precType->number; i++) {
	    if (pchoiceDev->papDevChoiceSet[i] != NULL) {
		pchoiceDev->papDevChoiceSet[i] = cvtOffsetToAddr(pchoiceDev,
		    pchoiceDev->papDevChoiceSet[i]);
	 	pchoiceDev->papDevChoiceSet[i]->papDevChoice = cvtOffsetToAddr(pchoiceDev,
		    pchoiceDev->papDevChoiceSet[i]->papDevChoice);
		for (j = 0; j < pchoiceDev->papDevChoiceSet[i]->number; j++) {
		    if (pchoiceDev->papDevChoiceSet[i]->papDevChoice[j] != NULL) {
			pchoiceDev->papDevChoiceSet[i]->papDevChoice[j] =
			    cvtOffsetToAddr(pchoiceDev,
			    pchoiceDev->papDevChoiceSet[i]->papDevChoice[j]);
			pchoiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice =
			    cvtOffsetToAddr(pchoiceDev,
			    pchoiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice);
		    }
		}
	    }
	}
	pdbBase->pchoiceDev = pchoiceDev;
    } else {
	/* ignore - already loaded */
	free(ptmp);
    }
    return 0;
}

/*
 * ADJ_DEVSUP
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_devSup(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    struct recDevSup *precDevSup = (struct recDevSup*)pdbBase->precDevSup;
    struct recType *precType = (struct recType*)pdbBase->precType;

    if (pdbBase->precType == NULL) {
	status = S_sdr_notLoaded;
	sprintf(text, "dbRead: adj_devSup() precType == NULL");
	errMessage(status, text);
	return (-1);
    }
    if (pdbBase->precDevSup == NULL) {
	precDevSup = (struct recDevSup *) ptmp;
	precDevSup->papDevSup = cvtOffsetToAddr(precDevSup,precDevSup->papDevSup);
	for (i = 0; i < precType->number; i++) {
	    if (precDevSup->papDevSup[i] != NULL) {
		precDevSup->papDevSup[i] = cvtOffsetToAddr(precDevSup,
			precDevSup->papDevSup[i]);
		precDevSup->papDevSup[i]->papDsetName = cvtOffsetToAddr(precDevSup,
			precDevSup->papDevSup[i]->papDsetName);
		precDevSup->papDevSup[i]->papDset = cvtOffsetToAddr(precDevSup,
			precDevSup->papDevSup[i]->papDset);
		for (j = 0; j < precDevSup->papDevSup[i]->number; j++) {
		    if (precDevSup->papDevSup[i]->papDsetName[j] != NULL) {
			precDevSup->papDevSup[i]->papDsetName[j] =
			    cvtOffsetToAddr(precDevSup,precDevSup->papDevSup[i]->papDsetName[j]);
		    }
		}
	    }
	}
	pdbBase->precDevSup = precDevSup;
    } else {
	/* ignore - already loaded */
	free(ptmp);
    }
    return 0;
}

/*
 * ADJ_CVTTABLE
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_cvtTable(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    struct arrBrkTable *pcvtTable = (struct arrBrkTable*) pdbBase->pcvtTable;

    if (pdbBase->pcvtTable == NULL) {
	pcvtTable = (struct arrBrkTable *) ptmp;
	pcvtTable->papBrkTable = cvtOffsetToAddr(pcvtTable,pcvtTable->papBrkTable);
	for (i = 0; i < pcvtTable->number; i++) {
	    if (pcvtTable->papBrkTable[i] != NULL) {
		pcvtTable->papBrkTable[i] = cvtOffsetToAddr(pcvtTable,
		    pcvtTable->papBrkTable[i]);
		pcvtTable->papBrkTable[i]->papBrkInt = cvtOffsetToAddr(pcvtTable,
		    pcvtTable->papBrkTable[i]->papBrkInt);
		pcvtTable->papBrkTable[i]->name = cvtOffsetToAddr(pcvtTable,
		    pcvtTable->papBrkTable[i]->name);
		for (j = 0; j < pcvtTable->papBrkTable[i]->number; j++) {
		    pcvtTable->papBrkTable[i]->papBrkInt[j] = cvtOffsetToAddr(pcvtTable,
			pcvtTable->papBrkTable[i]->papBrkInt[j]);
		}
	    }
	}
	pdbBase->pcvtTable = pcvtTable;
    } else {
	/* ignore - already loaded */
	free(ptmp);
    }
    return 0;
}

/*
 * ADJ_DRVSUP
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_drvSup(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    struct drvSup *pdrvSup = (struct drvSup*)pdbBase->pdrvSup;
    if (pdbBase->pdrvSup == NULL) {
	pdrvSup = (struct drvSup *) ptmp;
	pdrvSup->papDrvName = cvtOffsetToAddr(pdrvSup,pdrvSup->papDrvName);
	pdrvSup->papDrvet = cvtOffsetToAddr(pdrvSup,pdrvSup->papDrvet);
	for (i = 0; i < pdrvSup->number; i++) {
	    if (pdrvSup->papDrvName[i] != NULL)
		pdrvSup->papDrvName[i] = cvtOffsetToAddr(pdrvSup,pdrvSup->papDrvName[i]);
	}
	pdbBase->pdrvSup = pdrvSup;
    } else {
	/* ignore - already loaded */
	free(ptmp);
    }
    return 0;
}

/*
 * ADJ_RECSUP
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_recSup(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{

/* NOTE: this function never entered - iocInit creates rset's */
    status = S_sdr_noRecDef;
    sprintf(text, "dbRead: adj_recSup() not valid ");
    errMessage(status, text);
    return (-1);
}

/*
 * ADJ_SDRSUM
 *
 * Load SDR structure
 *
 */
static int
adj_sdrSum(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
    struct sdrSum  *psdrSum = (struct sdrSum*)pdbBase->psdrSum;
    if ( psdrSum == NULL ) {
	pdbBase->psdrSum = (struct sdrSum *) ptmp;
    } else {
	/* ignore - already loaded */
	free(ptmp);
    }
    return 0;
}

/*
 * ADJ_DBPVD
 *
 * Load and adjust SDR structure
 *
 */
static int
adj_dbPvd(ptmp, pdbBase)
    char           *ptmp;
    struct dbBase  *pdbBase;
{
/* keep this functionality for UNIX */
#ifdef UNIX
    if (pdbBase->pdbPvd == NULL) {
	pdbBase->pdbPvd = (struct pvd *) ptmp;
    } else {
	free(ptmp);
    }
#else
    /* In IOC skip pvd if present - this function will eventually go away */
    free(ptmp);
#endif
    return 0;
}

/*
 *
 */
static int
addValidDefaults(psdrFileSize, pSdrHeader)
    long           *psdrFileSize;
    struct sdrHeader *pSdrHeader;
{
/*NOTE: these must be kept in sync with sdrHeader.h for default.dctsdr structures*/
    switch (pSdrHeader->type) {
    case (SDR_DB_RECTYPE):
    case (SDR_DB_RECDES):
    case (SDR_CHOICEGBL):
    case (SDR_CHOICECVT):
    case (SDR_CHOICEREC):
    case (SDR_CHOICEDEV):
    case (SDR_DEVSUP):
    case (SDR_CVTTABLE):
    case (SDR_DRVSUP):
    case (SDR_RECSUP):		/* note: not an sdr type */
    case (SDR_ALLSUMS):
	break;
    default:
	return (0);
    }
    *psdrFileSize += pSdrHeader->nbytes + sizeof(struct sdrHeader);
    return (0);
}

/*
 * CHECKSDRSUM
 */
static int
checkSdrSum(psdrSum)
    struct sdrSum  *psdrSum;
{
#define MAX 128
    FILE    *fp;
    char            buff[MAX + 1];
    int             len;
    /* get the default.sdrSum */
    if ((fp = fopen("default.sdrSum", "r")) == 0) {
	errMessage(-1L, "FATAL ERROR FATAL ERROR FATAL ERROR FATAL ERROR");
	errMessage(-1L, "FATAL ERROR FATAL ERROR FATAL ERROR FATAL ERROR");
	errMessage(-1L, "\nCan't open default.sdrSum file.  Please invoke getrel\n");
	errMessage(-1L, "FATAL ERROR FATAL ERROR FATAL ERROR FATAL ERROR");
	errMessage(-1L, "FATAL ERROR FATAL ERROR FATAL ERROR FATAL ERROR");
	return (-1);
    }
    fgets(buff, MAX, fp);
    fclose(fp);

    len = strlen(psdrSum->allSdrSums);
    if ((strncmp(psdrSum->allSdrSums, buff, len)) != SAME) {
	errMessage(-1L, "FATAL ERROR FATAL ERROR FATAL ERROR FATAL ERROR");
	errMessage(-1L, "FATAL ERROR FATAL ERROR FATAL ERROR FATAL ERROR");
	sprintf(text, "\ndbRead: checkSdrSum - THIS DATABASE IS OUT_OF_DATE %s\n", current_file);
	errMessage(-1L, text);
	errMessage(-1L, "FATAL ERROR FATAL ERROR FATAL ERROR FATAL ERROR");
	errMessage(-1L, "FATAL ERROR FATAL ERROR FATAL ERROR FATAL ERROR");
	return (-1);
    }
    return 0;
}

/*
 * GETFILENAME
 * get next filename
 *	return 0/-1	success,empty
 */
static int getFileName(pfiles_line)
    char           pfiles_line[];
{
    int             len = 0;
    if (iptr == NULL)
	iptr = pfiles_line;
    if (*iptr == NULL) {
	iptr = NULL;
	return (-1);
    }
    /* kludge because the %n option of sscanf doesn't work properly on vxWorks !!! */
    current_file[0] = NULL;
    if ((sscanf(iptr, "%s", current_file)) != 1) {
	iptr = NULL;
	return (-1);
    }
    if ( (len = strlen(current_file)) == 0) {
	iptr = NULL;
	return (-1);
    }
    len = (len >= MAX_DBREAD_FILENAME) ? MAX_DBREAD_FILENAME-1 : len;
    iptr += len;
    /* move past leading spaces */
    while ( *iptr == ' ' )
	iptr++;
    return (0);
}

/*
 * INITDBREAD
 *
 * Inititialize dbBase and precHeader
 *
 */
static int
initDbRead(ppdbBase, pdatabaseName)
    struct dbBase **ppdbBase;
    char           *pdatabaseName;
{
    struct dbBase  *pdbBase;
    int             len;
    if ((*ppdbBase == NULL)) {
	if ((*ppdbBase = (struct dbBase *)
	  calloc((unsigned) 1, (unsigned) sizeof(struct dbBase))) == NULL) {
	    status = S_sdr_noAlloc;
	    sprintf(text, "dbRead: initDbRead() calloc error == NULL");
	    errMessage(status, text);
	    return (-1);
	}
    }
    pdbBase = *ppdbBase;
    /* clear the adj_fun array */
    for (i = 0; i < SDR_NTYPES; i++) {
	adj_fun[i] = 0;
    }
    /* init the adj func array */
    adj_fun[SDR_DB_RECTYPE] = adj_dbRecType;
    adj_fun[SDR_DB_RECORDS] = adj_dbRecords;
    adj_fun[SDR_DB_RECDES] = adj_dbRecDes;
    adj_fun[SDR_DB_PVD] = adj_dbPvd;
    adj_fun[SDR_CHOICEGBL] = adj_choiceGbl;
    adj_fun[SDR_CHOICECVT] = adj_choiceCvt;
    adj_fun[SDR_CHOICEREC] = adj_choiceRec;
    adj_fun[SDR_CHOICEDEV] = adj_choiceDev;
    adj_fun[SDR_DEVSUP] = adj_devSup;
    adj_fun[SDR_CVTTABLE] = adj_cvtTable;
    adj_fun[SDR_DRVSUP] = adj_drvSup;
    adj_fun[SDR_RECSUP] = adj_recSup;	/* Note: not an sdr type */
    adj_fun[SDR_ALLSUMS] = adj_sdrSum;
    /* save database name(s) */
    len = strlen(pdatabaseName);
    if (!pdbBase->pdbName) {
	if ((pdbBase->pdbName = (char *)
	     calloc((unsigned) 1, (unsigned) len + 1)) == NULL) {
	    status = S_sdr_noAlloc;
	    sprintf(text, "dbRead: initDbRead() calloc error == NULL");
	    errMessage(status, text);
	    return (-1);
	}
        strncpy((char *) pdbBase->pdbName, pdatabaseName, len + 1);
    }
    return (0);
}

/******************************************************************************
 * DBWRITE - writes the memory resident DB to a file
 * 
 * if databaseName exists - renames to databaseName% and then
 * 	creates databaseName
 * 
 * DEPENDS:	memory resident dbBase
 * 
 * returns		0,-1	OK,ERROR
 *
 ****************************************************************************/
#ifndef vxWorks /* currently not for vxWorks (doesn't support link call) */
#ifdef __STDC__
long dbWrite(struct dbBase *pdbBase, char *pdatabaseName)
#else
long dbWrite(pdbBase, pdatabaseName)
struct dbBase *pdbBase;
char *pdatabaseName;
#endif
{
    FILE           *fp_tmp;
    FILE           *fp_db;
    RECNODE        *precNode;
    ELLLIST           *preclist;
    int             pid;
    int             no_entries;
    unsigned        rec_size;	/* set to record size */
    struct stat     sbuf;
    struct sdrHeader sdrHeader;
    struct recLoc  *precLoc;
    short           recType;
    struct recHeader *precHeader = pdbBase->precHeader;
    struct recTypDes *pRecTypDes;	/* set to record description struct for this type */
    struct recDes  *precDes = pdbBase->precDes;
    char            tempfile[MAX_FILENAME_SZ + 1];
    char            residentdbname[MAX_FILENAME_SZ + 1];
    char            dbnamebak[MAX_FILENAME_SZ + 1];
    char           *iobuff;

    pid = getpid();

    sprintf(tempfile, "TMPdbWrite%d", pid);

    /* temp file is used initially for writing the new database */
    if ((fp_tmp = fopen(tempfile, "w")) == NULL) {
	sprintf(text, "dbWrite: can't fopen %s ", tempfile);
	errMessage(-1L, text);
	return (-1);
    }
    /* extract the 1st database name string from pdbBase->pdbName
     * so that the default sdr can be extracted from that file */

    /* form residentdbname from dbBase for default sdr copy */
    if ((sscanf(pdbBase->pdbName, "%s", residentdbname)) != 1) {
	sprintf(text, "dbWrite: sscanf - failed to get residentdbname from dbBase->pdbName");
	errMessage(-1L, text);
	return (-1);
    }
    /* copy the default sdr portion of the input database file to tempfile */
    /* Note: doing this is equivalent to dbCreate */
    if ((fp_db = fopen(residentdbname, "r")) == NULL) {
	fprintf(stderr, "dbWrite: can't fopen %s", residentdbname);
	return (-1);
    }
    if ((iobuff = (void *) calloc(1, (unsigned) pdbBase->sdrFileSize)) == NULL) {
	fprintf(stderr, "dbWrite: calloc failed");
	return (-1);
    }
    if ((fread(iobuff, (int) pdbBase->sdrFileSize, 1, fp_db)) != 1) {
	fprintf(stderr, "dbWrite: incomplete read of %s", residentdbname);
	fclose(fp_db);
	return (-1);
    }
    fclose(fp_db);
    if ((fwrite(iobuff, (int) pdbBase->sdrFileSize, 1, fp_tmp)) != 1) {
	fprintf(stderr, "dbWrite: incomplete write of %s", tempfile);
	fclose(fp_tmp);
	return (-1);
    }
    /*write (append) the memory resident database to the tempfile*/
    for (recType = 0; recType < precHeader->number; recType++) {
	if ((pRecTypDes = GET_PRECTYPDES(precDes, recType)) == NULL) continue;
	if ((precLoc = GET_PRECLOC(precHeader, recType)) == NULL) continue;
	if((preclist = precLoc->preclist) == NULL) continue;
	/* set up and write the SDR_DB_RECORDS sdrHeader */
	no_entries = ellCount(preclist);
	rec_size = pRecTypDes->rec_size;
	memset((char *) &sdrHeader, '\0', sizeof(struct sdrHeader));
	sdrHeader.magic = DBMAGIC;
	sdrHeader.nbytes = no_entries * rec_size + sizeof(struct recLoc);
	sdrHeader.type = SDR_DB_RECORDS;
	sdrHeader.create_date = (long) 0;
	if ((fwrite((char*)&sdrHeader, (int) sizeof(struct sdrHeader), 1, fp_tmp)) != 1) {
	    fprintf(stderr, "dbWrite: incomplete write of %s", tempfile);
	    fclose(fp_tmp);
	    return (-1);
	}
	if ((fwrite(precHeader->papRecLoc[recType],
		    (int) sizeof(struct recLoc), 1, fp_tmp)) != 1) {
	    fprintf(stderr, "dbWrite: incomplete write of %s", tempfile);
	    fclose(fp_tmp);
	    return (-1);
	}
	for (precNode = (RECNODE *) ellFirst(preclist);
		precNode != NULL;
		precNode = (RECNODE *) ellNext((ELLNODE*)precNode)) {
	    /* write  each record of this type contiguous */
	    if ((fwrite(precNode->precord, (int) rec_size, 1, fp_tmp)) != 1) {
		fprintf(stderr, "dbWrite: incomplete write of %s", tempfile);
		fclose(fp_tmp);
		return (-1);
	    }
	}
    }
    fclose(fp_tmp);
    /*here if database was written to tempfile*/
    strcpy(dbnamebak, pdatabaseName);
    strcat(dbnamebak, "%");
    /* if a backup is present - remove it*/
    if ((stat(dbnamebak, &sbuf)) == 0) {
	if (unlink(dbnamebak)) {
	    fprintf(stderr, "dbWrite: can't unlink %s", dbnamebak);
	    return (-1);
	}
    }
    /* if pdatabaseName file already exists - rename it to (pdatabaseName%) */
    if ((stat(pdatabaseName, &sbuf)) == 0) {
	if (link(pdatabaseName, dbnamebak)) {
	    fprintf(stderr, "dbWrite: can't link %s to %s", pdatabaseName, dbnamebak);
	    return (-1);
	}
	if (unlink(pdatabaseName)) {
	    fprintf(stderr, "dbWrite: can't unlink %s", pdatabaseName);
	    return (-1);
	}
    }
    /* rename tempfile to pdatabaseName */
    if (link(tempfile, pdatabaseName)) {
	fprintf(stderr, "dbWrite: can't link %s to %s", tempfile, pdatabaseName);
	return (-1);
    }
    if (unlink(tempfile)) {
	fprintf(stderr, "dbWrite: can't unlink %s", tempfile);
	return (-1);
    }
    return (0);
}
#endif /* not for vxWorks */




