static char SccsId[] = "$Id: dct.c,v 1.15 1995/02/28 22:04:42 mrk Exp $ ";

/*  $Id: dct.c,v 1.15 1995/02/28 22:04:42 mrk Exp $
 *	Original Author: Bod Dalesio (LANL)
 * 	Current Author : Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02  05-18-92        rcz     New database access (changed sdrLoad to dbRead)
 * 	...
 */

/*
 * Database Configuration Task
 *
 * This is the main routine for the Database Configuration Task
 * It contains the main menu which allows the user to:
 *		1. Report the Process Variables in the Control Network
 *		2. Create a Process Database
 *		3. Modify a Process Database
 *		4. Report a Process Database
 *		5. Delete a Process Database
 *		7. Select a report file or STDOUT
 *		8. Print a report file
 *		9. Modify from a text file
 *
 *
 *
 */
extern	char *version;		/* version string */	
#include	<stdlib.h>
#include	<stddef.h>
#include	<stdio.h>
#include	<curses.h>
#include	<ctype.h>
#include	<string.h>
#include	<errno.h>
#include	"dctmacros.h"
#include	<dbDefs.h>
#include	<dbRecords.h>
#include	<sdrHeader.h>
#include	<dbRecDes.h>	/* record/field definition structures */
#include	"dbPvd.h"	/* Process Variable Dir structure */
#include	"dctDef.h"

/*
 * Modules included in DCT are:
 * DCT.C
 *	MAIN		Database Configuration Task Mainline
 *	FC_READ		Reads in the Field Convert File
 *	FINITO		Closes all files when exitting DCT
 *	GET_REPORT_DEVICE	get the destination of reports
 *	INIT		Initializes globals and reads in the necessary files
 *	PVD_READ	Reads in the Process Variable Directory
 *	WRITE_MENU	Writes the mainline menu
 * COMSUBS.C	Library Subroutines
 *	BUFCMP		compare two n character buffers
 *	BUFCPY		copy an n character buffer to another
 *	CLEAR_ERROR	clears the error message portion of the screen
 *	ERROR_MSG	prints an error message on the screen
 *	FILL		fill an n character buffer with a specified character
 *	GETLINE1	get an input line, the first character is already read
 *	GETLINE		get an input line
 *	GET_FIELD	parse a field from a string
 *	GETINPCH	get an input character (converts ansii)
 *	INSTR		find a character in a string
 *	ISPRINTABL	returns true for a character which can be printed
 *	LEGAL_FLT_NUMBERS	check for legal floating point numbers
 *	LEGAL_HEX_NUMBERS	check for legal hexadecimal numbers
 *	LEGAL_NAME	check for a legal PV name
 *	LEGAL_NUMBERS	check for legal numbers for a long integer
 *	MENU_SELECT	gets a selection from a user for any menu
 *	NAME_CHAR	checks for a legal PV name character
 *	READLINE	read an input line from a disk file
 *	SPECIFIED_BY	interprets wildcards in strings
 *	TO_DOUBLE	convert string to float
 *	TO_HSHORT	convert string to short hexadecimal
 *	TO_SHORT	convert string to a short
 *	TO_VALUE	convert string to long integer
 * FILEDB.C	Subroutines that use a modifying file to change the database
 * 	FILEDB		top routine for file-to-database routines. Opens and
 *			reads modifying file.
 * 	FADD_PV		adds a process variable
 * 	FILE_STATUS	determines the status of a process variable
 * 	FMOD_PV		updates a field's value
 * 	GET_NODE_NUM	gets the node number of a node name
 * 	GET_TYPE_NUM	gets the type number from a type name
 * 	SPLIT_LINE	seperates a line into words and evaluates the words
 * 	STRIP_VAL	removes blanks from around a read-in value
 * PDSUBS.C	Subroutines that access the Process Database
 *	ADD_PD_NAME	Add a Process Database name to the Process Database list
 *	CREATE_PD	Creates a new Process Database
 *	DELETE_PD	Deletes all the Process Variables in a Process Database
 *	MACRO_MOD	Modifies specd field in specd PVs to specd value
 *	MODIFY_PD	Menu for performing operations on a Process Variables
 *	NODE_NUM_TO_NAME Converts a node number to a node name
 *	OPEN_PD		Opens a new Process Database
 *	PV_MENU		Displays the menu of operations for MODIFY_PD
 *	REPORT_PD	Reports the Process Database header & Process Variables
 *	RM_PD_NAME	Remove a process database name from the array
 *	SELECT_NODE	Displays the nodes and gets a selection from the user
 * PVD_SUB.C	Subroutines that access the Process Variable Directory
 *	ADDR_TO_TAG	Converts record type and field offset to PV name
 *	ADD_PV		Add a Process Variable to the directory
 *	DEL_PV		Delete a Process Variable from the directory
 *	FIND_SLOT	Find an available directory entry
 *	IN_NODE		Finds all PVs of the specified node and record type
 *	LIST_PVS	Print the Alphabetized Process Variable names
 *	ORDER_PVS	Alphabetizes the specified Process Variable names
 *	ORDER_SPECD_PVS	Alphabetizes the specified PV names (uses *'s in spec)
 *	PVD_DUMP	Dumps the Process Variable Directory binary (diagnostic)
 *	PVD_INIT	Initialize the Process Variable Directory
 *	PVD_LIST	Lists all Process Variable names in Alphbetical Order
 *	SAVE_PVD	Write the Process Variable Directory to disk
 * PV_SUBS.C	Subroutines that access a single Process Variable
 *	CLOSE_RECORD	close the currently open display and control record
 *	DUMP_RECORD	Dump a Process Variable record in binary (diagnostic)
 *	EXTEND_RECORD	Extends the database file to maintain # of spare records
 *	GET_AVL_PV	Find an available record in the Process Database
 *	OPEN_LOC_RECORD	open a record for local use
 *	OPEN_RECORD	open the display and control records for specified type
 *	PV_COPY		Copy a Process Variable to a new Process Variable
 *	PV_CREATE	Create a Process Variable
 *	PV_DELETE	Delete a Process Variable
 *	PV_MODIFY	Modify the parameters of a Process Variable
 *	PV_REPORT	Report a Process Variable
 *	READ_LOC_RECORD	read a record for local use (buffer and fd supplied)
 *	READ_RECORD	read the current database record into memory
 *	SELECT_A_PV	Get a PV selection from the Process Variables list
 *	SELECT_TYPE	Get a record type selection from available record types
 *	WRITE_LOC_RECOR	write a record from a specified buffer to a specified fd
 *	WRITE_RECORD	write the global record to disk
 * RDFSUBS.C	Subroutines that access the Record Definition File
 *	ADD_SP		convert field name to compare to stored field name
 *	CHANGE_FLDS	Go through the fields of a Process Variable for change
 *	CHANGE_RANGE	visually changes ranges
 *	CONVERT_FLD	Translates ASCII into a field value
 *	CRAM_FIELD	Place new value into specified PV and field name
 *	FLD_SEARCH	searches for field name and returns fields number
 *	FLD_TO_ASCII	Translates a field into ascii
 *	GET_CAMAC_ADDR	convert a str to a CAMAC address
 *	GET_MUX_CARD_TYPES gets the card types for each analog multiplexor card
 *	GET_RANGE_VAL	return value of a field for range change
 *	GET_VME_ADDR	convert a string to a VME address
 *	MOD_RECORD	Gets modifcations to each field of a Process Variable
 *	NEXT_VME_ADDR	gets the next available VME address
 *	NUMBER_OF_CHANNELS gets the number of channels of a specd card type
 *	NUMBER_OF_CARDS gets the number of cards of a specd card type
 *	PROMPT_FLD	determines whether a field should be prompted
 *	PUTFIELD	Places a new field value into the Process Variable
 *	ROTATE		Changes the value of a field with limited selection
 *	TO_FIELDNAME	Convert record type and offset to field name
 *	TYPE_IN_MEMORY	Gets the fields for the desired record type into memory
 *	WRITE_PAGE	Writes the fields of a Process Variable to the printer
 *	WRITE_PART	Refreshes screen after link type is changed
 *	VERIFY_VME_ADDR	Verifies a good VME address was specified
 *	XRECTYPE	Translates a record type name to a record type number
 * XREF.C	Subroutines that provide a cross reference of PV links
 *	CPY_REFS	Used by PV_COPY to fix all unresolved links in new PV
 *	DEL_REFS	Deletes all references to a specified PV
 *	LOC_SINGLE_REF	Provides a cross reference for a single PV in the 
 *			current node
 *	LOC_XREF	Provides a cross reference for the current node
 */
		
static void finito();

extern	char	*alloc_2d_array();
extern  void	exit();
/* file pointers */
int	pr_fd;		/* file pointer for the printer	*/
int	pvctrl_fd;	/* file pointer for the rec db file   */
int	rep_fd;		/* file pointer for the report device */


char	curr_dbname[PDNAME_SZ];	/* PD name in memory */
char	dctsdr_filename[MAX_FILENAME_SZ];   /* name of current .dctsdr file*/
char	pvd_filename[MAX_FILENAME_SZ];	/* name of current .pvd file */
char	rec_file[MAX_FILENAME_SZ];	/* name of current pd record file */

short	curr_rectype;		/* current record type in mem */
short	curr_rectypedef; 	/* current record type 	definitions pointed
					 to by the following */
struct	fldDes 	**pflddef;		/* current record's field
						definitions */
struct	recTypDes	*precorddes;		/* current record type 
						definition structure */
struct	recLoc		pd_recLoc;		/* header for a rec db file */
struct  sdrHeader	pd_sdrHeader;		/* sdr header for a rec dbfile*/


/* process variable */
char	ctrl_record[MAX_RECORD_SIZE];	/* PV being modified	*/

/* recently deleted process variable */
char	bu_record[MAX_RECORD_SIZE];
short	bu_type;		/* record type of recently deleted rec */

/* process databases defined */
char	*pd_names[MAXPDS];
char	*pdnamearray;		/* pointer to array containing pd name strings*/

/* report forms defined */
char	*frm_names[MAXFORMS];
char	*frmnamearray;		/* pointer to array containing form name
					 strings*/

int	dct_mode;		/* mode flag  0==interactive, 1==batch */
int	dis_range_check;	/* 0 = check if value within range, !=0 don't
					check */

/* report filename */
char	rep_filename[MAX_FILENAME_SZ];  /* report file name */
short	long_report;			/* flag for long report */

/* suppress error reporting flag */
short	suppress_errors;
short	suppress_errors1;	/* for non-curses errors */

/*********************************************/
struct dbBase *pdbBase;
struct choiceSet *choiceCvt;
struct arrChoiceSet *choiceGbl;
struct choiceRec *choiceRec;
struct devChoiceRec *choiceDev;
struct arrBrkTable *cvtTable;
struct recDes *dbRecDes;
struct recType *dbRecType;
struct recHeader *dbRecords;
struct recDevSup *devSup;
struct drvSup *drvSup;
struct recSup *recSup;
struct pvd *dbPvd;
/*********************************************/

/* the following sets up the main menu for dct */
/* the following routines need to be declared before the structures */
extern	int	pvd_list();
extern	int	modify_pd();
extern	int	delete_pd();
extern	int	report_unresolv();
extern	int	dct_utils();

int	create_pdx()
{
	return(create_pd(NULL));
}
int	report_pdx()
{
	return(report_pd(NULL));
}
int	summarize_pdx()
{
	return(summarize_pd(NULL,NULL));
}
int	filedbx()
{
	return(filedb(NULL,NULL));
}
int	exitx()
{
	finito();
	exit(0);
}

struct	menu_prompt main_menu_prompts[] = {
	0,pvd_list,	NOKEY, 	"List the Process Variables",
	1,create_pdx ,	NOKEY,	"Create a Process Database",
	2,report_pdx ,	NOKEY,	"Report a Process Database",
	3,modify_pd ,	NOKEY, 	"Open a Process Database",
	4,delete_pd ,	NOKEY, 	"Delete a Process Database",
	5,filedbx , 	NOKEY,	"Modify from a Text File",
	6,summarize_pdx,NOKEY,	"Print One-Line Summaries",
	7,report_unresolv,NOKEY,"Report Unresolved Process Variables",
	8,dct_utils,	NOKEY,	"DCT Utilities",
	9,exitx ,	EXIT,	"Exit DCT"
};

char	*title_header = "DCT version ";
char	*title_trailer = " (ANL/APS)";
char	title_string[80];
char	*mtstrings[] = {
	title_string,
	NULL
};

struct fmenu dct_main = {
	8,1,			/* title position */
	2,3,			/* place to start prompt text */
	0,3,			/* place to star prompt cursor */
	mtstrings,	/* array of title strings */
	sizeof(main_menu_prompts)/sizeof(struct menu_prompt), /* # of prompts */
	main_menu_prompts	/* the array  of prompts */
};


/*
 * MAIN
 *
 * database configuration task mainline
 */
main(argc, argv)
	int	argc;
	char	**argv;
{
	int	selected;		/* item of menu currently selected */
        pdbBase=NULL;	/* ptr to dbBase structure for dbRead */

	if ( argc > 1) 
		dct_mode = DCT_NONINTERACTIVE;	/* non interactive mode */
	else
		dct_mode = DCT_INTERACTIVE;	/* interactive mode	*/
	

	/* initialize global variables */
	if(init() < 0)
		 return(-1);

	if(dct_mode == DCT_NONINTERACTIVE)
		return(cmd_mode(argc,argv));

/* put up the menu and execute operator's choices */

	selected=0;
	while(1) {
		dis_range_check=0;
	     (*(dct_main.pmt[selected=fmenu(&dct_main, 1, selected)].proc))();
	}
}


/* finito
 *
 * finish up - close files
 */
static void finito(){

	/* close the report file descriptor */
	if (pr_fd > 0) close(pr_fd);
	
	FREE_MEM(pdnamearray);

	/* clear the screen */
	nocrmode();
	echo();
	refresh();
	endwin();
	
}
/*
 * GET_REPORT_DEVICE
 *
 * get the destination of reports
 */
get_report_device(){
	char		new_filename[MAX_FILENAME_SZ];
	short	new_pr_fd;

	memset(&new_filename[0],'\0',sizeof(new_filename));
	clear();
	move(0,0);
	printw("Filename to use for report: ");
	refresh();
	getline(&new_filename[0],80,0,strlen("Filename to use for report: "));
	if (!isalpha(new_filename[0])){
		if (pr_fd > 0) close(pr_fd);
		pr_fd = -1;
		rep_fd = STDOUT;
/*	}else if (strcmp(rep_filename,new_filename) == SAME){
		return;
*/
	}else if ((new_pr_fd = creat(new_filename,0666)) < 0){
		printf("Error %d creating %s",new_pr_fd,new_filename);
		getchar();
	}else{
		if (pr_fd > 0) close(pr_fd);
		pr_fd = new_pr_fd;
		rep_fd = pr_fd;
		strcpy(rep_filename,new_filename);
	}
}


/*
 * init
 *
 * initialize global variables
 */
init(){
char	*xname = "init";

	dis_range_check=0;
	curr_rectypedef = -1;
        curr_rectype = -1;
	bu_type = -1;

	memset(curr_dbname,'\0',PDNAME_SZ);
        memset(dctsdr_filename,'\0', sizeof(dctsdr_filename));
        memset(pvd_filename,'\0', sizeof(pvd_filename));
        memset(rec_file,'\0', sizeof(rec_file));


	/* init the process variable record fds */
	pvctrl_fd = -1;

	/* init the printer and report file descriptor */
	pr_fd = -1;
	rep_fd = STDOUT;

	if(dct_mode == DCT_NONINTERACTIVE)
		return(0);	/* non interactive mode */

	/* enable the cursor */
	initscr();
	noecho();
	crmode();

	if ( LINES < PAGE_LEN || COLS < PAGE_SIZE) {
		move(0,0);
		printw(
		"WARNING:  DCT requires a %d x %d (lines x cols) window size\n",
			PAGE_LEN, PAGE_WIDTH);
		printw("The current window size is only %d x %d\n", LINES,
			COLS);
		printw("You may encounter difficulty with some operations\n");
		printw("Please exit and increase your window size to %d x %d\n",				PAGE_LEN, PAGE_WIDTH);
		printw("If KP5 (EXIT) dosn't work, use Control E to exit\n");
		printw("Press any key to continue");
		refresh();
		getch();
	}
	/* enable error messages */
	suppress_errors = FALSE;
	suppress_errors1 = FALSE;

	if( (pdnamearray = 
	   alloc_2d_array( pd_names, MAXPDS, PDNAME_SZ)) == NULL) {
	       print_error_msg_p(xname, "Could not allocate database name array"
			,PAUSE_F);
		return(-1);
	}

	
	get_pdnames();  		/* get database names into pd_names[] */

	if( (frmnamearray = 
	   alloc_2d_array( frm_names, MAXFORMS, MAX_FILENAME_SZ)) == NULL) {
	       print_error_msg_p(xname, "Could not allocate form name array"
			,PAUSE_F);
		return(-1);
	}

	/* form title string */
	string_cat(title_string, sizeof(title_string), title_header, version,
			title_trailer);
	
	return(0);

}

/**************************************************************************
 *
 *  cmd_mode()
 *		runs dct in noniteractive mode
 *
 **************************************************************************/

int	cmd_mode(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    extern int	    opterr;

#define		REPORT	1
#define		CREATE  2
#define		MODIFY	3
#define		DELETE  4

    char	*xname="cmd_mode";
    int  	c;
    short 	rflg, cflg, dflg, mflg, iflg, oflg, zflg, errflg;
    int		command;
    char	database[MAX_FILENAME_SZ];
    char	mod_filename[MAX_FILENAME_SZ];	
    char	*def_reportfile= "short_report.";
    char	*usage = "Usage:\n\t%s \n\t%s -r <database_name> [ -o <report_name ]\n\t%s -d <database_name>\n\t%s -c <database_name>\n\t%s -m <database_name> [ -i <report_name>]\n";

    opterr = 0;		/* disable getopt() error reporting */
    rflg = cflg = dflg = mflg = iflg = oflg = zflg =  errflg = 0;
    while ((c = getopt(argc, argv, "r:d:c:m:o:i:")) != -1)
	switch (c) {
	case 'r':
		if (zflg || iflg)
			errflg++;
		else {	
			rflg++;
			zflg++;
			command = REPORT;
			strcpy(database, optarg);
			printf("database = %s\n", database);
		}
		break;
	case 'c':
		if (zflg || iflg || oflg)
			errflg++;
		else {	
			cflg++;
			zflg++;
			command = CREATE;
			strcpy(database, optarg);
			printf("database = %s\n", database);
		}
		break;
	case 'd':
		if (zflg || iflg || oflg)
			errflg++;
		else {	
			dflg++;
			zflg++;
			command = DELETE;
			strcpy(database, optarg);
			printf("database = %s\n", database);
		}
		break;
	case 'm':
		if (zflg || oflg)
			errflg++;
		else {	
			mflg++;
			zflg++;
			command = MODIFY;
			strcpy(database, optarg);
			printf("database = %s\n", database);
		}
		break;
	case 'o':
		if(mflg || cflg || dflg)
			errflg++;
		else {
			oflg++;
			strcpy(rep_filename, optarg);
			printf("report_file = %s\n", rep_filename);
		}
		break;
	case 'i':
		if(rflg || cflg || dflg)
			errflg++;
		else {
			iflg++;
			strcpy(mod_filename, optarg);
			printf("mod_file = %s\n", mod_filename);
		}
		break;
	default :
		errflg++;
		break;
	}
    if ( errflg || !(rflg || mflg || cflg || dflg)) {
	printf(usage, argv[0], argv[0], argv[0], argv[0], argv[0]);
	return(-1) ;
    }
    suppress_errors = TRUE;	/* turn off error reporting thru curses*/
    switch (command) {
	case REPORT:
		if(!oflg) {
			strcpy(rep_filename, def_reportfile);
			strcat(rep_filename, database);
		}
		if((rep_fd=creat(rep_filename,0666)) < 0) {
			print_syserr(xname, "Error creating report file",
					errno);
			printf("Report File Name = %s\n", rep_filename);
			return(-1);
		}
		return(report_pd(database));
		break;
	case CREATE:
		return(create_pd(database));
		break;
	case DELETE:
		return(delete_pd(database));
		break;
	case MODIFY:
		if(!iflg) {
			strcpy(mod_filename, def_reportfile);
			strcat(mod_filename, database);
		}
		return(filedb(database, mod_filename));
		break;
	default:		/* this should never happen */
		print_err(xname, "Unknown batch command detected\n");
		break;
    }
    return(0) ;
}

