/* pvd.c */
/* share/src/libCom $Id: pvd.c,v 1.9 1995/08/17 14:44:40 mrk Exp $ */

/* dbPvd.c
 *  former name was dbcon/src/CONVERTSUBS.C
 *
 * record name and field name convert subroutines
 *
 *      Author:          Bob Dalesio
 *      Date:            3-18-88
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	12-17-88	lrd	added choicera process in ret_field
 * .03	05-01-89	lrd	fixed ret_pv (assumed inputs would be less then
 *				PVNAME_SZ length)
 * .04	05-12-89	lrd	optimized conversion routines
 * .05	05-12-89	lrd	reworked the hash algorithm for a better
 *				distribution
 * .06	05-18-89	lrd	choice processing gets 0 when no match
 * .07  03-21-90	mrk	changed name to dbPvd
 *				made changes to support new dbPvd.h
 * .08	07-17-91	rac	don't compile get_recfld for EPICS_V2
 * .09	05-17-92	rcz	moved pvd.c from libCom to dct
 * .10  05-18-92        rcz     New database access 
 * .11  08-11-93	mrk	remove references to EPICS_V2
 * .12  05-12-95        adc     Little endian support (esp. Alpha/OSF-1)
 *
 */
#include	<dbDefs.h>
#include	"dbPvd.h"
#include	<dbRecDes.h>

#if defined(EPICS_LITTLE_ENDIAN)
#define BYTE_SWAP_4(A)            \
  ( (((A) & 0xFF) << 24 )    |    \
    (((A) & 0xFF00) << 8)    |    \
    (((A) & 0xFF0000) >> 8)  |    \
    (((A) & 0xFF000000) >> 24) )
#endif


extern struct pvd *dbPvd;
extern struct recDes *dbRecDes;
/* Forward reference for pvdGetFld */
struct fldDes *pvdGetFld();

/* function declarations */
static bufcpy();


/* global variables */

/*
 * HASH NOTES:
 *
 * 1. Databases need to be reported and restored when these are changed !!!!!!
 *
 * 2. The goal is to maximize the number of unique indices into the hash area. 
 *	A characteristic of a good hash function is a good distribution
 *	over the range of the hash table when the table is sparsely filled.
 *
 * 3. These mulitpliers were determined by trial and error. These produced the
 *	best results for the 5 databases which existed at the time.
 *
 * 4. The define PVNAME_SZ was 28 and the hash table size was 512 when this 
 *	table was built.
 */
static short	hash_shift[PVNAME_SZ] =  {	/* modifiers */
					/* all prime numbers- I think */
103,1,101,2,97,3,89,5,83,7,79,11,73,13,71,17,67,19,61,23,59,29,53,31,47,37,43,41};

/* END OF DISASTER AREA */

/*
 * Subroutines
 *
 * get_recfld	convert a record type and field name to offset, size, field type
 *      args
 *		type		record type
 *		pfldname	pointer to the field name
 *		ptype		pointer to the field type
 *		poffset		pointer to the field offset
 *		psize		pointer to the field size
 *	returns
 *		0		found
 *		-1		not found
 *
 * pvdHash		get a collision chain index for a PV name
 *	args
 *		str		pointer to the Process Variable name
 *	returns
 *		hval		the collision chain index
 * pvdPvSearch 	find the specified PV name in the PVD
 *	args
 *		name		pointer to the Process Variable name
 *		hashinx		the collision chain index
 *	returns
 *		>= 0		PVD entry index of the specified PV
 *		-1		not found
 * pvdGetFld	fetches the field type etc.
 *      args
 *		type		record type
 *		pfldname	pointer to field name
 *	returns
 *		NULL		not found
 *		struct fldDes*	If found
 * pvdGetRec	fetches the TYPE and RECORD_NO of a Process Variable
 *	arguments
 *		ppv_name	pointer to the name of the Process Variable
 *		type		pointer to the record type of the PV
 *		record_no	record number of this PV
 *	outputs
 *		-1		record not defined
 *		0		record found
 */

/* GET_RECFLD - */

int get_recfld(type,pfldname,ptype,poffset,psize)
short	type;		/*record type*/
char	*pfldname;	/*field name */
short	*ptype;		/*field type */
short	*poffset;	/*field offset*/
short	*psize;		/*field size */
{
	struct fldDes	*pfldDes;

	if(!(pfldDes=pvdGetFld(type,pfldname))) return(-1);
	*ptype = pfldDes->field_type;
	*poffset = pfldDes->offset;
	*psize = pfldDes->size;
	return(0);
}


/*
 *	HASH
 *
 *	hash a process variable name
 */
short pvdHash(str)
register char *str;
{
	register short	i;
	register short	hval;
	register short	*pval = &hash_shift[0];

	hval = 0;
	for (i = 0; i < PVNAME_SZ && *str; i++){
		hval = hval + (*(unsigned char *)str++ * *pval++);
	}
	hval &= HASH_NO -1;	/* truncate value */

	return(hval);
}

/*
 *	PV_SEARCH
 *
 *	Attempt to find a given process variable
 */
short pvdPvSearch(pname,hashinx)
register char	*pname;
register short	hashinx;
{
	register short	i,inx;
	register char	*ptbl;

	/* chain thru hash table entries, testing for equality */
	i = dbPvd->hash_entry[hashinx];
	while (i != NO_ENTRY){
		ptbl = &dbPvd->entry[i].name[0];
		inx = 0;
		while((inx < PVNAME_SZ) 
		  && (*(long *)(pname+inx) == *(long *)(ptbl+inx)))
			inx += sizeof(long); 
		if (inx == PVNAME_SZ)
			return(i);
		i = dbPvd->entry[i].next;
	}
	return(-1);
}

/*
 * RET_FIELD
 *
 * get a field from a record
 */
struct fldDes *pvdGetFld(type,pfldname)
short			type;		/* record type		*/
char			*pfldname;	/* field name		*/
{
	short			i;
	struct fldDes		*pfield;
	struct recTypDes	*precTypDes;
	register short		top,bottom,test;
	char			fieldname[FLDNAME_SZ];
	register char		*pfieldname = &fieldname[0];
	register unsigned long	*sortFldName;
	register short		*sortFldInd;

	/* verify that pvd and recDes are available	*/
	if( !dbPvd || !dbRecDes ) return(NULL);
	/* verify that type is valid			*/
	if((type<0) || (type>=dbRecDes->number)) return(NULL);
	/* get and check pointer to this record's field definitions */
	precTypDes=dbRecDes->papRecTypDes[type];
	if(!precTypDes) return(NULL);
	sortFldName = precTypDes->sortFldName;
	sortFldInd  = precTypDes->sortFldInd;

	/* blank pad the field name */
	i = 0;
	while(*(pfldname+i) && (i < FLDNAME_SZ)){
		*(pfieldname+i) = *(pfldname+i);
		i++;
	}
	while(i < FLDNAME_SZ){
		*(pfieldname+i) = ' ';
		i++;
	}
        
	/* check for default */
	if (*pfieldname == ' ')
		bufcpy(pfieldname,"VAL ",FLDNAME_SZ);

	/* binary search through ordered field names */
	top = precTypDes->no_fields -1;
	bottom = 0;
	test = (top + bottom) / 2;
	while(1){
	        unsigned int t1,t2;
		t1 = (unsigned int) sortFldName[ test];
		t2 = (unsigned int) *(unsigned int*)pfieldname;

#if defined(EPICS_LITTLE_ENDIAN)
		t1 = BYTE_SWAP_4( t1);
		t2 = BYTE_SWAP_4( t2);
#endif

		/* check the field name */
		if ( t1 == t2) {
			pfield = precTypDes->papFldDes[sortFldInd[test]];
			return(pfield);
		}else if ( t1 > t2) {
			top = test - 1;
			if (top < bottom) return(NULL);
			test = (top + bottom) / 2;
		}else{
			bottom = test + 1;
			if (top < bottom) return(NULL);
			test = (top + bottom) / 2;
		}
	}
}

/*
 *	RET_PV
 *
 *	return the  type and record number of a process variable from 
 *	the directory
 */
int pvdGetRec(ppv_name,ptype,precord_no)
register char		*ppv_name;
short			*ptype;
short			*precord_no;
{

	register short	i,index;
	register short	hshval;
	char		record_name[PVNAME_SZ];

	/* check that the process variable directory is present */
	if (!dbPvd) return(-2);

	/* get the hash pointer */
	hshval = pvdHash(ppv_name);

	/* find the process variable in the directory */
	bufcpy(record_name,ppv_name,
	  ((strlen(ppv_name) > PVNAME_SZ)?PVNAME_SZ:strlen(ppv_name)) );
	for (i = strlen(ppv_name); i < PVNAME_SZ; i++) record_name[i] = 0;
	if ((index = pvdPvSearch(record_name,hshval)) < 0)
		return(-1);

	/* assign the type and offset */
	*precord_no = dbPvd->entry[index].record_no;
	*ptype = dbPvd->entry[index].record_type;

	return(0);
}

static bufcmp(pstr1,pstr2,length)
register char   *pstr1;
register char   *pstr2;
register short  length;
{
        register short  i;

        for(i=0; i<length; i++){
                if(*(pstr1+i) > *(pstr2+i)) return(1);
                if(*(pstr2+i) > *(pstr1+i)) return(-1);
        }

        return(0);
}

static bufcpy(pstr1,pstr2,length)
register char   *pstr1;
register char   *pstr2;
register short  length;
{
        register short  i;

        for (i=0; i<length; i++)
                *(pstr1+i) = *(pstr2+i);
        return(i);
}
